"""DataFrame mutable object and column type"""

from ...app import settings

from sqlalchemy.types import JSON, TypeDecorator
from sqlalchemy_mutable import MutableList, MutableDict

import csv
import json
from copy import copy
from datetime import datetime
from io import StringIO


class Variable(MutableList):
    """
    Store the data for a single variable of the dataframe.

    Attributes
    ----------
    fill_rows : bool
        Indicates that the variable should be padded with the last entry in 
        this variable. If `False`, the variable will be padded with `None`. 
        This is `True` if the `rows` parameter in the `add` method is negative 
        and `False` otherwise.
    """
    def __init__(self, fill_rows=False, source=[]):
        # set _python_type to None to avoid validation for setattr
        self._python_type = None
        self.fill_rows = fill_rows
        super().__init__(source)
        
    def add(self, entry, rows=1):
        """
        Add an entry (or list of entries) to the variable.
        
        Parameters
        ----------
        entry : literal or list
            Entry or entries to add to the variable.

        rows : int, default=1
            Number of rows in which this entry should appear. If rows is negative, this indicates that the variable should be padded with the last entry in this variable.

        Returns
        -------
        self
        """
        def convert(item):
            return str(item) if isinstance(item, datetime) else item

        if isinstance(entry, list):
            self += [convert(item) for item in entry]
        else:
            self += [convert(entry)]
        self.fill_rows = rows < 0
        return self
    
    def pad(self, end_row):
        """Pad the entries
        
        Add padding so that the Variable length is equal to `end_row`.
        """
        val = self[-1] if self and self.fill_rows else None
        self += [val]*(end_row-len(self))
        return self


class DataFrame(MutableDict):
    """
    Dataframe for storing participant data.

    Attributes
    ----------
    filename : str
        Name of the download file associated with this dataframe.
    """
    def __init__(self, source={}):
        # set _python_type to None to avoid validation in setattr
        self._python_type = None
        self.filename = None
        super().__init__(
            {key: Variable(source=value) for key, value in source.items()}
        )

    def max_row(self, variables=None):
        """
        Maximum number of rows of any of the given variables.

        Parameters
        ----------
        variables : list of str or None, default=None
            List of variable names (keys of self). If `None`, this method 
            returns the maximum row of any of the variables in the dataframe.

        Returns
        -------
        rows : int
            Maximum number of rows of any of the given variables.
        """
        variables = variables or self.keys()
        lengths = [len(self[var]) for var in variables if var in self]
        return max(lengths) if lengths else 0
        
    def append(self, data):
        """
        Append a dataframe to the current dataframe.

        Parameters
        ----------
        data : dict
            Data to append to the current dataframe

        Returns
        -------
        self
        """
        self.pad()
        self.add(data)
        self.pad()
        return self

    def add(self, data, rows=1):
        """
        Parameters
        ----------
        data : dict
            Data to add to the current dataframe; maps variable names to an 
            entry or list of entries.

        rows : int, default=1
            Number of rows the data should occupy if the entries are not 
            lists. If the entries are lists, rows will not matter.

        Returns
        -------
        self
        """
        self._changed()
        start_row = self.max_row(data.keys())
        for var, entry in data.items():
            self.prep_variable(var, start_row)
            self[var].add(entry, rows)
        return self
        
    def prep_variable(self, var, end_row=None):
        """
        Prepare a variable before adding an entry.

        Parameters
        ----------
        var : str
            Name of the variable (in `self.keys()`) to prepare.

        end_row : int or None, default=None
            The expected end row of the variable. If the variable does not yet contain enough rows, it will be padded. If `None`, the end row is assumed to be the max row of the current dataframe.

        Returns
        -------
        self
        """
        end_row = self.max_row() if end_row is None else end_row
        if var not in self:
            self[var] = Variable()
        self[var].pad(end_row)
        return self
    
    def remove(self, start, end):
        """Remove data between start and end indices"""
        self._changed()
        for var in self.keys():
            modified_data = self[var][:start] + self[var][end:]
            self[var] = Variable(self[var].fill_rows)
            self[var].add(modified_data)
    
    def pad(self):
        """Pad DataFrame so all Variables have the same number of rows"""
        self._changed()
        end_row = self.max_row()
        [self[var].pad(end_row) for var in self.keys()]

    def get_download_file(self):
        """Get file download tuple
        
        File download is (filename, file string) tuple.
        """
        data = self.copy()
        if not settings['collect_IP']:
            data.pop('IPv4', None)
        csv_str = StringIO()
        writer = csv.writer(csv_str)
        writer.writerow(data.keys())
        writer.writerows(zip(*data.values()))
        return self.filename, csv_str

    def dump(self):
        return json.dumps(dict(filename=self.filename, data=self))

    @classmethod
    def load(cls, state_dict):
        state = json.loads(state_dict)
        df = cls(state['data'])
        df.filename = state['filename']
        return df


class DataFrameType(TypeDecorator):
    impl = JSON

    def process_bind_param(self, value, dialect):
        return value.dump()

    def process_result_value(self, value, dialect):
        return DataFrame.load(value)


DataFrame.associate_with(DataFrameType)


"""Data download"""

from ...app import bp, db
from ...models import Page, Participant
from ...qpolymorphs import Check, Download, Input, Select
from ...models.private import DataStore
from ...tools import chromedriver, join, show_on_event
from .login import researcher_login_required
from .utils import navbar

from bs4 import BeautifulSoup
from docx import Document
from docx.shared import Inches
from flask import request, session

import os
import re
import time
from base64 import b64encode
from datetime import timedelta
from io import BytesIO
from itertools import chain
from zipfile import ZipFile

# Screen buffer for driver window height 
HEIGHT_BUFFER = 74
# Width of survey view for docx
SURVEY_VIEW_IMG_WIDTH = Inches(6)
# number of seconds for <iframe> to load before taking screenshot
IFRAME_LOAD_TIME = 3

"""1. Create a download page"""
@bp.route('/download', methods=['GET','POST'])
@researcher_login_required
def download():
    page = get_download_page()
    db.session.commit()
    session['download-page-id'] = page.id
    return page._render()

def get_download_page():
    if 'download-page-id' in session:
        download_page = Page.query.get(session['download-page-id'])
        if download_page:
            return download_page
    
    part_ids_q = Input(
        "Enter participant IDs for survey viewing.",
        validate=valid_part_ids
    )
    presentation_q = Select(
        "Select presentation for viewing.",
        [('First presentation only', 'first'), ('All presentations', 'all')],
    )
    show_on_event(presentation_q, part_ids_q, '\S+', regex=True)
    file_type_q = Select(
        "Select the type of download.",
        [('Screenshots', 'screenshot'), ('Text', 'text')],
    )
    show_on_event(file_type_q, part_ids_q, '\S+', regex=True)

    page = Page(
        Check(
            "Select files to download.", 
            [('Data frame', 'data'), ('Participant metadata', 'meta')],
            multiple=True,
        ),
        part_ids_q,
        presentation_q,
        file_type_q,
        Download(
            callback=request.url, 
            download_msg='Download complete',
            handle_form_functions=handle_download_form,
        ),
        navbar=navbar.render(), back=False, forward=False
    )
    db.session.add(page)
    return page

def valid_part_ids(survey_view_q):
    """
    Validate that requested participant IDs correspond to participants in 
    the database.

    Parameters
    ----------
    survey_view_q : hemlock.Input
        Input question where researchers input participant IDs.

    Returns
    -------
    error_msg : str or None
        A message displaying invalid participant IDs. If this message is 
        not `None`, the download will stop. Otherwise, it proceeds.
    """
    part_ids = re.split(r'\W+', survey_view_q.response)
    part_ids = [id for id in part_ids if id]
    invalid_ids = [i for i in part_ids if Participant.query.get(i) is None]
    if invalid_ids:
        return "<p>The following participant ID was invalid: {}.</p>".format(
            join('and', *invalid_ids)
        )

def handle_download_form(response, btn):
    """
    Adds the appropriate file creation function to the download button.

    Parameters
    ----------
    response : dict
        Form response; not used.

    btn : hemlock.Download
        Download button (last question on the download page).
    """
    page = get_download_page()
    btn.create_file_functions.clear()
    btn.downloads.clear()
    page._record_response()
    if page._validate():
        btn.create_file_functions = FileCreator(
            dataframes=page.questions[0].response,
            part_ids=page.questions[1].response,
            presentation=page.questions[2].response,
            download_type=page.questions[3].response
        )


class FileCreator():
    """
    Creates the download zip file.

    Parameters
    ----------
    dataframes : dict
        Maps dataframe name (`'data'` or `'meta'`) to bool indicating that the
        researcher requested this dataframe.

    part_ids : list
        List of participant IDs requested for survey view.

    presentation : str
        Which page presentation to deliver for the survey view; first 
        presentation only or all presentations.

    download_type : str
        Type of download for the survey view; screenshot or text.

    Attributes
    ----------
    btn : hemlock.Download
    
    datastore : hemlock.models.DataStore

    driver : selenium.webdriver.chrome.webdriver.WebDriver
        Driver for taking screenshots for survey view.

    download_type : str
        Set from parameter.
    
    files : list of (filename, str [bytes]) tuples
        Requested download files.

    presentation : str
        Set from paramaeter.

    requested_dataframes : dict
        Set from `dataframes` parameter.

    requested_part : list of hemlock.Participant
        Participants whose survey view the researcher requested.
    """
    def __init__(self, dataframes, part_ids, presentation, download_type):
        self.btn = None
        self.datastore = None
        self.driver = None
        self.files = []

        self.requested_dataframes = dataframes
        part_ids = re.split(r'\W+', part_ids)
        part_ids = [id for id in part_ids if id]
        self.requested_parts = Participant.query.filter(
            Participant.id.in_(part_ids)
        ).all()
        self.presentation = presentation
        self.download_type = download_type

    def __call__(self, btn):
        """
        Create and deliver the zip file with requested data files.

        Parameters
        ----------
        btn : hemlock.Download
        """
        # note: need to add requested participants before adding `to_store`
        # otherwise you get an error that you try to add the same participant
        # to the session multiple times.
        # I don't know why changing the order in which you add these lists
        # fixes the problem.
        db.session.add_all(self.requested_parts)
        self.btn = btn
        self.datastore = DataStore.query.first()
        if 'data' in self.requested_dataframes:
            for expr in self.prep_dataframe():
                yield expr
        if 'meta' in self.requested_dataframes:
            for expr in self.prep_meta():
                yield expr
        if self.requested_parts:
            for expr in self.gen_views():
                yield expr
        for expr in self.zip_files():
            yield expr

    def prep_dataframe(self):
        """
        Prepare the main data frame for download. Begin by storing data from
        participants whose data was updated since they were last stored in the
        data frame, or whose data has not yet been stored.
        """
        def store_updated():
            # stores ids of participants whose data was flagged as updated
            stage = 'Storing participants with updates'
            yield self.btn.reset(stage, 0)
            updated = [part for part in parts if part.updated]
            for i, part in enumerate(updated):
                if i % 10 == 0:
                    # commit every 10 participants to avoid memory issues
                    db.session.commit()
                yield self.btn.report(stage, 100.*i/len(updated))
                self.datastore.store_participant(part)
            yield self.btn.report(stage, 100)

        def store_missing():
            """
            Check for participants whose data are missing from the data 
            frame. Why might this happen? Imagine two participants complete 
            the survey in close succession. The first queries the data 
            store, but before he commits his data, the second queries the 
            data store. The first commits his data, but this commit is 
            overridden when the second participant commits his data.
            """
            stage = 'Collecting participant data'
            yield self.btn.reset(stage, 0)
            stored_ids = list(set(self.datastore.data.get('ID') or []))
            stored_ids.sort()
            stored_ids_head = parts_head = 0
            while parts_head < len(parts):
                if parts_head % 10 == 1:
                    # commit every 10 participants to avoid memory issues
                    db.session.commit()
                yield self.btn.report(stage, 100.*parts_head/len(parts))
                # Note: id of the participant at index parts_head is 
                # parts_head + 1
                if (
                    stored_ids_head < len(stored_ids) 
                    and stored_ids[stored_ids_head] == parts_head + 1
                ):
                    stored_ids_head += 1
                else:
                    self.datastore.store_participant(parts[parts_head])
                parts_head += 1
            yield self.btn.report(stage, 100)

        parts = Participant.query.order_by('id').all()
        for expr in store_updated():
            yield expr
        for expr in store_missing():
            yield expr
        self.files.append(self.datastore.data.get_download_file())
        yield self.btn.report('Data frame prepared', 100)

    def prep_meta(self):
        """
        Prepare the participant metadata file.
        """
        stage = 'Preparing participant metadata'
        yield self.btn.reset(stage, 0)
        self.files.append(self.datastore.meta.get_download_file())
        yield self.btn.report(stage, 100)

    def gen_views(self):
        """
        Generate survey views for all requested participants.
        """
        if self.download_type == 'screenshot':
            self.driver = chromedriver(headless=True)
        generator = chain.from_iterable(
            (self.gen_view(part) for part in self.requested_parts)
        )
        for event in generator:
            yield event
        if self.driver is not None:
            self.driver.close()
            self.driver = None

    def gen_view(self, part):
        """
        Generate survey view for a single participant.

        Parameters
        ----------
        part : hemlock.Participant
        """
        stage = 'Creating survey view for participant '+str(part.id)
        yield self.btn.reset(stage, 0)
        doc = Document()
        if self.presentation == 'all':
            pages = part._viewing_pages
        elif self.presentation == 'first':
            pages = [p for p in part._viewing_pages if p.first_presentation]
        for i, page in enumerate(pages):
            yield self.btn.report(stage, 100.*i/len(pages))
            self.add_page_to_doc(doc, page)
        doc_bytes = BytesIO()
        doc.save(doc_bytes)
        self.files.append(('Participant-{}.docx'.format(part.id), doc_bytes))

    def add_page_to_doc(self, doc, page):
        """
        Add a viewing page to the document.

        Parameters
        ----------
        doc : docx.document.Document
            Survey view document for the current participant.

        page : hemlock.models.private.ViewingPage
            Page to add to the document.
        """
        def write_screenshot_to_doc():
            dist = os.environ.get('WSL_DISTRIBUTION')
            self.driver.get('file://'+('wsl$/'+dist+path if dist else path))
            self.accept_alerts()
            width = self.driver.get_window_size()['width']
            # width = 1024
            height = self.driver.execute_script(
                'return document.body.parentNode.scrollHeight'
            )
            form = self.driver.find_element_by_tag_name('form')
            try:
                _ = form.find_element_by_tag_name('iframe')
                time.sleep(IFRAME_LOAD_TIME)  # give the frame time to load
                height = max(height, 768)
            except:
                pass
            self.driver.set_window_size(width, height+HEIGHT_BUFFER)
            with BytesIO() as page_bytes:
                page_bytes = BytesIO()
                page_bytes.write(form.screenshot_as_png)
                doc.add_picture(page_bytes, width=SURVEY_VIEW_IMG_WIDTH)

        if self.driver is None:
            text = BeautifulSoup(page.html, 'html.parser').text.strip('\n')
            doc.add_paragraph(text)
            return
        path = page.mkstmp()
        try:
            write_screenshot_to_doc()
        except:
            pass
        os.remove(path)

    def accept_alerts(self):
        """
        Accept any alerts which may prevent the driver from loading the page 
        html.
        """
        try:
            while True:
                self.driver.switch_to.alert.accept()
        except:
            pass

    def zip_files(self):
        """
        Put all requested files in a zip archive and store in the download
        button's temporary downloads list.
        """
        stage = 'Zipping files'
        yield self.btn.reset(stage, 0)
        with BytesIO() as zipf_bytes:
            zipf = ZipFile(zipf_bytes, 'w')
            for i, (filename, io) in enumerate(self.files):
                yield self.btn.report(stage, 100.*i/len(self.files))
                zipf.writestr(filename, io.getvalue())
                io.close()
            zipf.close()
            data = b64encode(zipf_bytes.getvalue()).decode()
            url = 'data:application/zip;base64,' + data
            self.btn.tmp_downloads = [(url, 'hemlock-survey-data.zip')]
        yield self.btn.report(stage, 100)


"""# Dashboard"""

from .. import tools
from ..app import db
from ..models import Question
from ..tools import iframe, key

from flask import render_template
from sqlalchemy_mutable import MutableDictType, MutableDictJSONType

import re
from urllib.parse import parse_qs, urlencode, urlparse


class Dashboard(Question):
    """
    Embeds a <a href="https://plotly.com/dash/" target="_blank">dash app</a>.

    Parameters
    ----------
    label : str or None, default=None
        Dashboard label.

    template : str, default='hemlock/dash.html'
        Template for the dashboard body.

    Attributes
    ----------
    src : str
        `src` attribute of the `<iframe>` tag.

    Examples
    --------
    In this example, we create a simple dash app in which participants click on a
    button. We embed this app in a hemlock survey, and record the number of times
    a participant clicked the button.

    Install `dash` with:

    ```bash
    $ hlk install dash
    ```

    Or:

    ```bash
    $ pip install dash
    ```

    In `survey.py`:

    ```python
    from hemlock import Branch, Dashboard, Label, Page, route

    @route('/survey')
    def start():
    \    return Branch(
    \        Page(
    \            Dashboard(src='/dashapp/', var='n_clicks')
    \        ),
    \        Page(
    \            Label('The end.'),
    \            terminal=True
    \        )
    \    )
    ```

    In `app.py`:

    ```python
    import survey

    import dash
    import dash_core_components as dcc
    import dash_html_components as html
    from dash.dependencies import Input, Output
    from hemlock import Dashboard, create_app

    app = create_app()
    dash_app = dash.Dash(
    \    server=app,
    \    routes_pathname_prefix='/dashapp/',
    \    external_stylesheets=['https://codepen.io/chriddyp/pen/bWLwgP.css']
    )

    dash_app.layout = html.Div([
    \    dcc.Location(id='url'),
    \    html.Button('Click me!', id='button'),
    \    html.P(id='click-tracker'),
    ])

    @dash_app.callback(
    \    Output('click-tracker', 'children'),
    \    [Input('url', 'search'), Input('button', 'n_clicks')]
    )
    def update_clicks(search, n_clicks):
    \    n_clicks = n_clicks or 0
    \    Dashboard.record_response(search, n_clicks)
    \    return '{} clicks'.format(n_clicks)

    if __name__ == '__main__':
    \    from hemlock.app import socketio
    \    socketio.run(app, debug=True)
    ```

    Run the app with:

    ```bash
    $ hlk serve
    ```

    Or:

    ```bash
    $ python app.py
    ```

    Then open your browser and navigate to <http://localhost:5000/>.
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'dashboard'}

    g = db.Column(MutableDictType)
    iframe_kwargs = db.Column(MutableDictJSONType)
    security_key = db.Column(db.String)

    _iframe_kwargs_keys = [
        'src', 'aspect_ratio', 'query_string', 'div_class', 'div_attrs',
        'iframe_class', 'iframe_attrs'
    ]

    def __init__(self, label=None, template='hemlock/dash.html', **kwargs):
        self.iframe_kwargs = {}
        self.security_key = key()
        super().__init__(label=label, template=template, **kwargs)

    def __getattribute__(self, key):
        if key == '_iframe_kwargs_keys' or key not in self._iframe_kwargs_keys:
            return super().__getattribute__(key)
        return self.iframe_kwargs.get(key)

    def __setattr__(self, key, val):
        if key in self._iframe_kwargs_keys:
            self.iframe_kwargs[key] = val
        else:
            super().__setattr__(key, val)

    @classmethod
    def get(cls, search):
        """
        Utility for retrieving a dashboard question in a dash callback.

        Parameters
        ----------
        search : str, formatted as URL query string
            Must have 'id' and 'key' parameters.

        Returns
        -------
        dashboard : hemlock_dash.Dashboard
            Dashboard specified by the id in the search string.

        Examples
        --------
        ```python
        ...
        import dash_core_components as dcc
        from hemlock import Dashboard

        app.layout = html.Div([
        \    dcc.Location(id='url', refresh=False),
        \    ...
        ])

        @app.callback(
        \    Output(...),
        \    [Input('url', 'search'), ...]
        )
        def my_callback(search, ...):
        \    dashboard = Dashboard.get(search)
        ```
        """
        qs = parse_qs(urlparse(search).query)
        id, key = qs['id'][0], qs['security_key'][0]
        dash = cls.query.get(id)
        return dash if dash.security_key == key else None

    @classmethod
    def record_response(cls, search, response):
        """
        Utility for writing the `response` attribute of the dashboard 
        question.

        Parameters
        ----------
        search : str, formatted as URL query string
            Must have 'id' and 'key' parameters.

        response : 
            Value to which to set the dash question's `repsonse` attribute.

        Returns
        -------
        dashboard : hemlock_dash.Dashboard
            Dashboard specified by the id in the search string.

        Examples
        --------
        ```python
        ...
        import dash_core_components as dcc
        from hemlock import Dashboard

        app.layout = html.Div([
        \    dcc.Location(id='url', refresh=False),
        \    ...
        ])

        @app.callback(
        \    Output(...),
        \    [Input('url', 'search'), ...]
        )
        def my_callback(search, ...):
        \    Dashboard.record_response(search, 'hello world')
        """
        dashboard = cls.get(search)
        dashboard.response = response
        db.session.commit()
        return dashboard

    def render_url(self):
        qs = self.iframe_kwargs.get('query_string', {}).copy()
        qs.update({'id': self.id, 'security_key': self.security_key})
        return self.src + '?' + urlencode(qs)

    def _render(self):
        # add id and security key to query string
        src = self.render_url()
        kwargs = {
            key: val for key, val in self.iframe_kwargs.items() 
            if key not in ('src', 'query_string')
        }
        return render_template(
            self.template, q=self, embed=iframe(src, **kwargs)
        )

    def _record_response(self):
        # response should be recorded in a callback
        return self


"""# Questions

`hemlock.Question` and `hemlock.ChoiceQuestion` are 'question skeletons';
most useful when fleshed out. See section on question polymorphs.
"""

from ..app import db
from ..tools import key
from .bases import Data
from .private import CSSListType, JSListType

from flask import render_template, request
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import validates
from sqlalchemy_mutable import (
    HTMLAttrsType, MutableType, MutableJSONType, MutableListType, 
    MutableListJSONType
)

import html
import os
from copy import copy


class Question(Data):
    """
    Base object for questions. Questions are displayed on their page in index 
    order.

    It inherits from 
    [`hemlock.models.Data`](bases.md).

    Parameters
    ----------
    label : str or None, default=None
        Question label.

    template : str, default='form-group.html'
        File name of question template.

    extra_css : list, default=[]
        List of extra CSS elements to add to the default CSS.

    extra_js : list, default=[]
        List of extra javascript elements to add to the default 
        javascript.

    form_group_class : list, default=['card', 'form-group', 'question']
        List of form group classes.

    form_group_attrs : dict, default={}
        Dictionary of HTML attribues for the form group tag.

    error_attrs : dict, default={'class' ['alert', 'alert-danger']}
        Dictionary of HTML attributes for the error alert.

    Attributes
    ----------
    append : str or None, default=None
        Text (usually) appended to the input tag.

    css : list, default=[]
        List of CSS elements.

    default : misc
        Default question response.

    error : str or None, default=None
        Error message.

    error_attrs : dict
        Set from the `error_attrs` parameter.

    form_group_class : list
        Set from the `form_group_class` parameter.

    form_group_attrs : dict
        Set from the `form_group_attrs` parameter.

    has_responded : bool, default=False
        Indicates that the participant has responded to this question.

    input_attrs : dict
        Dictionary of HTML attributes for the input tag.

    label : str or None, default=None
        Question label.

    prepend : str or None, default=None
        Text (usually) prepended to the input tag.

    response : misc
        Participant's response.

    template : str
        Set from the `template` parameter.

    Relationships
    -------------
    part : hemlock.Participant or None
        The participant to which this question belongs. Derived from 
        `self.page`.

    branch : hemlock.Branch or None
        The branch to which this question belongs. Derived from `self.page`.

    page : hemlock.Page or None
        The page to which this question belongs.

    compile : list of hemlock.Compile, default=[]
        List of compile functions; run before the question is rendered.

    validate : list of hemlock.Validate, default=[]
        List of validate functions; run to validate the participant's response.

    submit : list of hemlock.Submit, default=[]
        List of submit functions; run after the participant's responses have been validated for all questions on a page.

    debug : list of hemlock.Debug, default=[]
        List of debug functions; run during debugging. The default debug function is unique to the question type.

    Notes
    -----
    A CSS element can be any of the following:

    1. Link tag (str) e.g., `'<link rel="stylesheet" href="https://my-css-url">'`
    2. Href (str) e.g., `"https://my-css-url"`
    3. Style dictionary (dict) e.g., `{'body': {'background': 'coral'}}`

    The style dictionary maps a css selector to an attributes dictionary. The attributes dictionary maps attribute names to values.

    A Javascript element can be any of the following:

    1. Attributes dictionary (dict) e.g., `{'src': 'https://my-js-url'}`
    2. JS code (str)
    3. Tuple of (attributes dictionary, js code)
    """
    id = db.Column(db.Integer, db.ForeignKey('data.id'), primary_key=True)
    question_type = db.Column(db.String)
    __mapper_args__ = {
        'polymorphic_identity': 'question',
        'polymorphic_on': question_type
    }

    # relationships
    @property
    def part(self):
        return None if self.page is None else self.page.part

    @property
    def branch(self):
        return None if self.page is None else self.page.branch

    _page_id = db.Column(db.Integer, db.ForeignKey('page.id'))

    # HTML attributes
    key = db.Column(db.String(10))
    template = db.Column(db.String)
    css = db.Column(CSSListType, default=[])
    js = db.Column(JSListType, default=[])
    form_group_class = db.Column(MutableListJSONType, default=[])
    form_group_attrs = db.Column(HTMLAttrsType, default={})
    error = db.Column(db.Text)
    error_attrs = db.Column(HTMLAttrsType, default={})
    label = db.Column(db.Text)
    prepend = db.Column(db.String)
    append = db.Column(db.String)
    input_attrs = db.Column(HTMLAttrsType)

    # Function attributes
    compile = db.Column(MutableListType, default=[])
    debug = db.Column(MutableListType, default=[])
    validate = db.Column(MutableListType, default=[])
    submit = db.Column(MutableListType, default=[])

    # Additional attributes
    default = db.Column(MutableJSONType)
    response = db.Column(MutableJSONType)
    has_responded = db.Column(db.Boolean, default=False)

    def __init__(
            self, label=None, extra_css=[], extra_js=[], 
            form_group_class=['card', 'form-group', 'question'],
            form_group_attrs={}, 
            error_attrs={'class': ['alert', 'alert-danger']}, 
            **kwargs
        ):
        def add_extra(attr, extra):
            # add extra css or javascript
            if extra:
                assert isinstance(extra, (str, list))
                attr += [extra] if isinstance(extra, str) else extra

        self.key = key(10)
        self.compile, self.debug, self.validate, self.submit = [], [], [], []
        self.css, self.js = [], []
        super().__init__(
            label=label, 
            form_group_class=form_group_class,
            form_group_attrs=form_group_attrs,
            error_attrs=error_attrs, 
            **kwargs
        )
        add_extra(self.css, extra_css)
        add_extra(self.js, extra_js)

    # methods
    def clear_error(self):
        """
        Clear the error message.

        Returns
        -------
        self : hemlock.Question
        """
        self.error = None
        return self

    def clear_response(self):
        """
        Clear the response.

        Returns
        -------
        self : hemlock.Question
        """
        self.response = None
        self.has_responded = False
        return self

    # methods executed during study
    def _compile(self):
        [f(self) for f in self.compile]
        return self

    def _render(self, body=None):
        return render_template(self.template, q=self)

    def _render_js(self):
        return self.js.render()

    def _record_response(self):
        self.has_responded = True
        self.response = request.form.get(self.key)
        if isinstance(self.response, str):
            # convert to safe html
            self.response = html.escape(self.response)
        return self
        
    def _validate(self):
        """Validate Participant response
        
        Check validate functions one at a time. If any yields an error 
        message (i.e. error is not None), indicate the response was invalid 
        and return False. Otherwise, return True.
        """
        for f in self.validate:
            self.error = f(self)
            if self.error:
                return False
        self.error = None
        return True

    def _record_data(self):
        self.data = self.response
        return self
    
    def _submit(self):
        [f(self) for f in self.submit]
        return self

    def _debug(self, driver):
        [f(driver, self) for f in self.debug]
        return self


class ChoiceQuestion(Question):
    """
    A question which contains choices. Inherits from `hemlock.Question`.

    Parameters
    ----------
    label : str or None, default=None
        Question label.

    choices : list, default=[]
        Choices which belong to this question. List items are usually 
        `hemlock.Choice` or `hemlock.Option`.

    template : str or None, default=None
        Template for the question body.

    Attributes
    ----------
    choices : list, default=[]
        Set from `choices` parameter.

    choice_cls : class, default=hemlock.Choice
        Class of the choices in the `choices` list.

    multiple : bool, default=False
        Indicates that the participant can select multiple choices.

    Notes
    -----
    A choice can be any of the following:

    1. Choice objects (type will depend on the choice question).
    2. `str`, treated as the choice label.
    3. `(choice label, value)` tuple.
    4. `(choice label, value, name)` tuple.
    5. Dictionary with choice keyword arguments.
    """
    choices = None # must be implemented by choice question

    def __init__(self, label=None, choices=[], template=None, **kwargs):
        self.choices = choices
        super().__init__(label=label, template=template, **kwargs)

    def _render(self, body=None):
        """Add choice HTML to `body`"""
        return render_template(self.template, q=self)

    def _record_response(self):
        """Record response

        The response is a single choice or a list of choices (if multiple 
        choices are allowed).
        """
        self.has_responded = True
        idx = request.form.getlist(self.key)
        self.response = [self.choices[int(i)].value for i in idx]
        if not self.multiple:
            self.response = self.response[0] if self.response else None
        return self

    def _record_data(self):
        """Record data

        For single choice questions, the data is the selected choice's 
        `value`.

        For multiple choice questions, the data is a dictionary mapping 
        each choice's `value` to a binary indicator that it was selected.
        """
        if self.multiple:
            self.data = {
                choice.value: int(choice.value in self.response)
                for choice in self.choices
            }
        else:
            self.data = self.response
        return self
    
    def _pack_data(self):
        """Pack data for storage in the `DataStore`

        For multiple choice questions, the packed data dictionary is similar to the data, but with the question's variable prepended to the key.
        """
        var = self.var
        if not self.multiple or var is None:
            return super()._pack_data()
        if self.data is None:
            packed_data = {
                var+c.value: None 
                for c in self.choices if c.value is not None
            }
        elif isinstance(self.data, dict):
            packed_data = {var+key: val for key, val in self.data.items()}
        else:
            packed_data = self.data
        return super()._pack_data(packed_data)


"""Router models

This file defines two router models. 

Each participant has a main Router. The Router has two 'tracks'. When the 
participant requests a new page, the router moves through the 'request' 
track. The request track compiles and renders a new survey page. When the 
participant submits a page, the router moves through the 'submit' track. The 
submit track records the participant's response, validates it, submits it, 
navigates, and redirects with a new page request.

The secondary Navigator router is nested in the main Router. It handles 
backward and forward navigation.
"""

from ...app import db
from .viewing_page import ViewingPage

from flask import current_app, request, redirect, session, url_for
from flask_worker import RouterMixin, set_route

from datetime import datetime


class Router(RouterMixin, db.Model):
    """
    The main router belongs to a Participant. It handles the request and 
    submit tracks.

    Parameters
    ----------
    gen_root : callable
        Callable which generates the root branch.

    Attributes
    ----------
    part : hemlock.Participant
        Participant to whom this router belongs.

    page : hemlock.Page
        The participant's current page.

    in_progress : bool
        Indicates that a route is being processed. Helps avoid queueing 
        extra routing functions.

    job_in_progress : bool
        Indicates that a worker's job is being processed.

    view_function : str
        Name of the function which generated the root branch. It is also the 
        name of the view function associated with this router. On redirect, 
        the router, redirects to `url_for(self.view_function)`.

    navigator : hemlock.models.private.Navigator
        A navigator router which handles navigation between pages.
    """
    id = db.Column(db.Integer, primary_key=True)
    part_id = db.Column(db.Integer, db.ForeignKey('participant.id'))
    in_progress = db.Column(db.Boolean, default=False)
    view_function = db.Column(db.String)
    worker_page = db.Column(db.Text)
    worker = db.relationship('Worker', uselist=False)
    navigator = db.relationship('Navigator', backref='router', uselist=False)

    @property
    def page(self):
        return self.part.current_page

    def __init__(self, gen_root):
        self.view_function = gen_root.__name__
        self.navigator = Navigator(gen_root)
        super().__init__(self.compile)

    def __call__(self, *args, **kwargs):
        """Route overload

        If the participant's time has expired, render the current page with 
        a time expired error message.

        The router starts at the beginning of the request track (compile). If 
        the participant has just submitted a page, the router moves to the 
        beginning of the submit track (record response). Then route as normal.

        Returns
        -------
        page : str (html)
            HTML of the participant's next page, or a loading page.
        """
        def render_current_page():
            if self.page.timer and self.page.timer.state != 'running':
                self.page.timer.start()
                db.session.commit()
            return self.page._render()

        def process_request():
            # commit the in_progress flag so future requests will know to 
            # return the loading page
            self.in_progress = True
            db.session.commit()
            if request.method == 'POST' and self.func.name == 'compile':
                # participant has just submitted the page
                self.func = self.record_response
            result = self.func(self, *args, **kwargs)
            self.in_progress = False
            db.session.commit()
            return result

        if self.part.time_expired:
            self.page.error = current_app.time_expired_text
            return self.page._render()
        if self.worker:
            return (
                process_request() if self.worker.job_finished 
                else self.worker_page
            )
        if self.in_progress:
            # send loading page while the next page loads
            return current_app.settings.get('loading_page')
        if request.method == 'GET' and not self.page.run_compile:
            # page refresh or participant went back to this page
            return render_current_page()
        return process_request()

    """Request track"""
    @set_route
    def compile(self):
        return self.run_worker(
            self.page, '_compile', self.page.compile_worker, self.render
        )
    
    def render(self):
        if self.page.terminal and not self.part.completed:
            self.part.end_time = datetime.utcnow()
            self.part.completed = True
        page_html = self.page._render()
        self.part._viewing_pages.append(
            ViewingPage(page_html, first_presentation=not self.page.viewed)
        )
        self.page.viewed = True
        if self.page.timer is not None:
            self.page.timer.start()
        return page_html
    
    """Submit track"""
    def record_response(self):
        if self.page.timer is not None:
            self.page.timer.pause()
        self.part.end_time = datetime.utcnow()
        self.part.completed, self.part.updated = False, True
        self.page._record_response()
        if self.page.direction_from == 'back':
            if self.page.compile_worker:
                self.page.compile_worker.reset()
            self.navigator.back(self.page.back_to)
            return self.redirect()
        return self.validate()
    
    @set_route
    def validate(self):
        if not current_app.settings['validate']:
            # validation may be off for testing purposes
            return self.submit()
        return self.run_worker(
            self.page, '_validate', self.page.validate_worker, self.submit
        )
    
    @set_route
    def submit(self):
        if not self.page.is_valid() and current_app.settings['validate']:
            # will redirect to the same page with error messages
            return self.redirect()
        return self.run_worker(
            self.page, '_submit', self.page.submit_worker, self.forward_prep,
        )

    def forward_prep(self):
        """Prepare for forward navigation
        
        Check if direction_from has been changed by submit functions. If so, 
        navigate appropriately. Otherwise, reset the navigator in preparation 
        for forward navigation.
        """
        page = self.page
        if page.direction_from == 'back':
            self.navigator.back(page.back_to)
        if page.direction_from in ('back', 'invalid') or page.last_page():
            return self.redirect()
        self.navigator.reset()
        return self.forward(page.forward_to)
    
    @set_route
    def forward(self, forward_to):
        """
        Forward navigation uses the Navigator subrouter. If the navigator 
        has not yet started forward navigation (i.e. `not nav.in_progress`), 
        the main router gets a loading page from `nav.forward`. Otherwise, it 
        gets a loading page from `nav.__call__`.

        The navigator may need a worker to generate a new branch. While the 
        worker's job is in progress, the navigator returns a loading page 
        (which is not `None`).

        When it finds the next survey page, the navigator return `None`
        loading page. However, if there is a specified page to which the 
        router is navigating forward (`forward_to is not None`), it may be 
        necessary to call `nav.forward` again.
        
        When the navigator has finished, the router redirects the 
        participant to the new survey page.

        Parameters
        ----------
        forward_to : hemlock.Page
            Page to which to navigate.
        """
        loading_page = (
            self.navigator.func(self.navigator) if self.navigator.in_progress
            else self.navigator.forward(forward_to)
        )
        if loading_page is None and forward_to is not None:
            loading_page = self.navigator.forward(forward_to)
        return loading_page or self.redirect()

    def redirect(self):
        self.reset()
        url_arg = 'hemlock.{}'.format(self.view_function)
        if self.part.meta.get('Test') is None:
            return redirect(url_for(url_arg))
        # during testing, you may have multiple users running at once
        # in this case, participants much be found be ID, not `current_user`
        return redirect(
            url_for(url_arg, part_id=self.part.id, part_key=self.part._key)
        )

    def run_worker(
            self, obj, method_name, worker, next_route, *args, **kwargs
        ):
        """
        Run a worker if applicable.

        Parameters
        ----------
        obj : Page or Branch
            Object whose method should be run.

        method_name : str
            Name of the object's method to run. e.g. '_compile'

        worker : hemlock.WorkerMixin or None
            Worker which executes the method, if applicable.

        next_route : router method
            The next route to navigate to after executing the method.

        \*args, \*\*kwargs :
            Arguments and keyword arguments for `next_route`.

        Returns
        -------
        page : str (html)
            Page for the client. This will be a loading page, if the worker is 
            running, or the next page of the survey.        
        """
        if worker is None:
            getattr(obj, method_name)()
            return next_route(*args, **kwargs)
        if worker.job_finished:
            worker.reset()
            self.worker, self.worker_page = None, None
            return next_route(*args, **kwargs)
        self.worker = worker
        self.worker_page = worker.enqueue_method(obj, method_name)
        return self.worker_page


class Navigator(RouterMixin, db.Model):
    """    
    The navigator subrouter is nexted in the main router. It handles forward 
    and backward navigation.

    Attributes
    ----------
    router : hemlock.models.private.Router
        Main router with which the navigator is associated.

    in_progress : bool, default=False
        Indicates that navigation is in progress.
    """
    id = db.Column(db.Integer, primary_key=True)
    router_id = db.Column(db.Integer, db.ForeignKey('router.id'))
    in_progress = db.Column(db.Boolean, default=False)

    @property
    def part(self):
        return self.router.part

    @property
    def branch_stack(self):
        return self.part.branch_stack

    @property
    def branch(self):
        return self.part.current_branch

    @branch.setter
    def branch(self, value):
        self.part.current_branch = value

    @property
    def page(self):
        return self.part.current_page

    def reset(self):
        super().reset()
        self.in_progress = False

    """Forward navigation"""
    @set_route
    def forward(self, forward_to=None):
        """Advance forward to specified Page"""
        self.in_progress = True
        if forward_to is None:
            return self.forward_one()
        while self.page != forward_to:
            loading_page = self.forward_one()
            if loading_page is not None:
                return loading_page
    
    def forward_one(self):
        """Advance forward one page"""
        if self.page._eligible_to_insert_branch():
            return self.insert_branch(self.page)
        self.branch._forward()
        return self.forward_recurse()
    
    @set_route
    def insert_branch(self, origin):
        """Grow and insert new branch into the branch stack"""
        # note the parallel between this function and Router.run_worker
        worker = origin.navigate_worker
        if worker is None:
            origin._navigate()
            return self._insert_branch(origin)
        if worker.job_finished:
            worker.reset()
            self.router.worker, self.router.worker_page = None, None
            return self._insert_branch(origin)
        self.router.worker = worker
        self.router.worker_page = worker.enqueue_method(origin, '_navigate')
        return self.router.worker_page
    
    def _insert_branch(self, origin):
        branch = origin.next_branch
        self.branch_stack.insert(self.branch.index+1, branch)
        self.increment_head()
        return self.forward_recurse()

    def forward_recurse(self):
        """Recursive forward function

        Advance forward until the next Page is found (i.e. is not None).
        """
        if self.page is not None:
            return
        if self.branch._eligible_to_insert_branch():
            return self.insert_branch(self.branch)
        self.decrement_head()
        self.branch._forward()
        return self.forward_recurse()
    
    """Back navigation"""
    def back(self, back_to=None):
        """Navigate backward to specified Page"""
        if back_to is None:
            return self.back_one()
        while self.page != back_to:
            self.back_one()
            
    def back_one(self):
        """Navigate backward one Page"""
        # re-run compile functions the next time the participant goes forward 
        # to this page
        self.page.run_compile = True
        if self.page == self.branch.start_page:
            return self.remove_branch()
        self.branch._back()
        return self.back_recurse()
        
    def remove_branch(self):
        """Remove current branch from the branch stack"""
        self.decrement_head()
        self.branch_stack.pop(self.branch.index+1)
        return self.back_recurse()
        
    def back_recurse(self):
        """Recursive back function
        
        Navigate backward until previous Page is found.
        """
        if self.found_previous_page():
            return
        if self.page is None:
            if self.branch.next_branch in self.branch_stack:
                self.increment_head()
            elif not self.branch.pages:
                return self.remove_branch()
            else:
                self.branch._back()
        else:
            self.increment_head()
        return self.back_recurse()
    
    def found_previous_page(self):
        """Indicate that previous page has been found in backward navigation
        
        The previous page has been found when 1) the Page is not None and
        2) it does not branch off to another Branch in the stack.
        """
        return (
            self.page is not None 
            and self.page.next_branch not in self.branch_stack
        )

    """Move the head of the branch stack"""
    def increment_head(self):
        self.branch = self.branch_stack[self.branch.index+1]
    
    def decrement_head(self):
        self.branch = self.branch_stack[self.branch.index-1]


"""# Common bases and mixins"""

from ..app import db
from .functions import Compile, Debug, Validate, Submit, Navigate

from bs4 import BeautifulSoup
from flask import current_app, render_template
from sqlalchemy import Column, inspect
from sqlalchemy_modelid import ModelIdBase
from sqlalchemy_mutable import (
    HTMLAttrsType, MutableType, MutableJSONType, MutableListType, 
    MutableListJSONType, MutableModelBase
)

import os


class Base(ModelIdBase):
    """
    Base for all Hemlock models.

    Interits from  
    [`sqlalchemy_modelid.ModelIdBase`](https://dsbowen.github.io/sqlalchemy-modelid/).

    Parameters
    ----------
    \*\*kwargs :
        You can set any attribute by passing it as a keyword argument.

    Attributes
    ----------
    name : str or None, default=None
        Used primarily as a filter for database querying.
    """
    def __init__(self, **kwargs):
        settings = current_app.settings.get(type(self).__name__)
        settings = settings.copy() if settings else {}
        settings.update(kwargs)
        [setattr(self, key, val) for key, val in settings.items()]
        super().__init__()


class BranchingBase(Base):
    """
    Attributes
    ----------
    navigate : callabe or None, default=None
        A callable which returns the next branch.
    """
    navigate = db.Column(MutableType)

    def _eligible_to_insert_branch(self):
        """Indicate that object is eligible to grow and insert next branch
        
        A Page or Branch is eligible to insert the next branch to the
        Participant's branch_stack iff the navigator is not None and
        the next branch is not already in the branch_stack.
        """
        return (
            self.navigate is not None 
            and self.next_branch not in self.part.branch_stack
        )

    def _navigate(self):
        from .branch import Branch

        def set_relationships():
            self.next_branch = next_branch
            if isinstance(self, Branch):
                next_branch.origin_branch = self
                next_branch.origin_page = None
            else:
                # self is hemlock.Page
                next_branch.origin_branch = None
                next_branch.origin_page = self

        next_branch = self.navigate(self)
        assert isinstance(next_branch, Branch)
        set_relationships()
        next_branch.current_page = next_branch.start_page
        return self


class Data(Base, MutableModelBase, db.Model):
    """
    Polymorphic base for all objects which contribute data to the dataframe.

    Data elements 'pack' their data and return it to their participant, who in turn sends it to the data store.

    Attributes
    ----------
    data : sqlalchemy_mutable.MutableType, default=None
        Data this element contributes to the dataframe.

    data_rows : int, default=1
        Number of rows this data element contributes to the dataframe for its 
        participant. If negative, this data element will 'fill in' any emtpy
        rows at the end of the dataframe with its most recent value.

    index : int or None, default=None
        Order in which this data element appears in its parent; usually a 
        `hemlock.Branch`, `hemlock.Page`, or `hemlock.Question`.

    var : str or None, default=None
        Variable name associated with this data element. If `None`, the data 
        will not be recorded.

    record_order : bool, default=False
        Indicates that the order of this data element should be recorded in
        the datafame. The order is the order in which this element appeared
        relative to other elements with the same variable name.

    record_index : bool, default=False
        Indicates that the index of this data element should be recorded in
        the dataframe. The index is the order in which this element appeared
        relative to other elements with the same parent. For example, the
        index of a question is the order in which the question appeared on its
        page.

    record_choice_index : bool, default=False
        Indicates that the index of this data element's choices should be 
        recorded in the dataframe. For example, a `hemlock.Check` question has
        multiple choices that the participant can select. The index of a 
        choice is its index in the question's choice list.
    """
    id = db.Column(db.Integer, primary_key=True)
    data_type = db.Column(db.String)
    __mapper_args__ = {
        'polymorphic_identity': 'data',
        'polymorphic_on': data_type
    }

    data = db.Column(MutableType)
    data_rows = db.Column(db.Integer, default=1)
    index = db.Column(db.Integer)
    var = db.Column(db.Text)
    record_order = db.Column(db.Boolean, default=False)
    record_index = db.Column(db.Boolean, default=False)
    record_choice_index = db.Column(db.Boolean, default=False)

    def _pack_data(self, data=None):
        """
        Pack data for storing in data store.

        Parameters
        ----------
        data : dict or None, default=None
            Partly packed data from a downstream class.

        Returns
        -------
        packed_data : dict
        
        Notes
        -----
        `var`Index is the index of the object; its order within its
        Branch, Page, or Question. `var`Order is the order of the data element
        relative to other data elements with the same variable.
        """
        if self.var is None:
            return {}
        if data is None:
            data = {self.var: None if self.data is None else self.data}
        if hasattr(self, 'order') and self.record_order:
            data[self.var+'Order'] = self.order
        if hasattr(self, 'index') and self.record_index:
            data[self.var+'Index'] = self.index
        if hasattr(self, 'choices') and self.record_choice_index:
            data.update({
                self.var+c.name+'Index': idx
                for idx, c in enumerate(self.choices) if c.name is not None
            })
        return data


"""# Label"""

from ..app import db, settings
from ..models import Question

settings['Label'] = {'data': 1}

class Label(Question):
    """
    This question contains a label and does not receive input from the 
    participant.

    Parameters
    ----------
    label : str or None, default=None
        Question label.

    template : str, default='hemlock/label.html'
        Path to the Jinja template for the label body.

    Examples
    --------
    ```python
    from hemlock import Label, Page, push_app_context

    app = push_app_context()

    Page(Label('Hello World')).preview()
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'label'}

    def __init__(self, label=None, template='hemlock/label.html', **kwargs):
        super().__init__(label=label, template=template, **kwargs)

    def _record_data(self):
        return self


"""# Participant"""

from ..app import db, settings
from ..tools import key
from .bases import Base
from .private import DataFrame, DataStore, Router

from flask_login import UserMixin, login_user
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy_mutable import MutableDictType

from datetime import datetime
from random import randint

settings['Participant'] = {'g': {}, 'meta': {}}

def _send_data(func):
    """Send data to the DataStore
    
    Begin by recording the Participant's previous status. If the status has
    changed as a result of the update, register the change in the `DataStore`.
    """
    def status_update(part, update):
        part._previous_status = part.status
        return_val = func(part, update)
        if part._previous_status == part.status:
            # status hasn't changed; no need to update data store
            return return_val
        if part.status in ('Completed', 'TimedOut'):
            # cache the participant's data before querying the data store
            part.get_data()
        ds = DataStore.query.first()
        ds.update_status(part)
        if part.status in ('Completed', 'TimedOut'):
            ds.store_participant(part)
        db.session.commit()
        return return_val

    return status_update


class Participant(UserMixin, Base, db.Model):
    """
    The Participant class stores data for an individual survey participant and
    handles navigation for that participant.

    Inherits from [`hemlock.models.Base`](bases.md).

    Attributes
    ----------
    completed : bool, default=False
        Indicates that the participant has completed the survey.

    end_time : datetime.datetime
        Last time the participant submitted a page.

    g : dict, default={}
        Dictionary of miscellaneous objects.

    meta : dict, default={}
        Participant metadata, such as IP address.

    start_time : datetime.datetime
        Time at which the participant started he survey.

    status : str, default='InProgress'
        Participant's current status; `'InProgress'`, `'TimedOut'`, or 
        `'Completed'`. Read only; derived from `self.completed` and 
        `self.time_expired`.

    time_expired : bool, default=False
        Indicates that the participant has exceeded their allotted time for 
        the survey.

    updated : bool, default=True
        Indicates that the participant's data was updated after the last time 
        their data was stored; if `True`, the participant's data will be 
        re-stored when data are downloaded.

    Relationships
    -------------
    branch_stack : list of hemlock.Branch
        The participant's stack of branches.

    current_branch : hemlock.Branch
        Participant's current branch (head of `self.branch_stack`).

    current_page : hemlock.Page
        Participant's current page (head of `self.current_branch`).

    pages : list of hemlock.Page
        Pages belonging to the participant.

    embedded : list of hemlock.Embedded
        Embedded data elements belonging to the participant.

    data_elements : list of hemlock.DataElement
        List of all data elements belonging to the participant, ordered by 
        `id`.

    Examples
    --------
    ```python
    from hemlock import Branch, Label, Page, Participant, push_app_context

    def start():
    \    return Branch(Page(Label('<p>Hello World</p>')))

    app = push_app_context()

    part = Participant.gen_test_participant(start)
    part.current_page.preview()
    ```
    """
    id = db.Column(db.Integer, primary_key=True)
    _key = db.Column(db.String(90))

    _router = db.relationship(
        'Router', 
        backref='part', 
        uselist=False
    )
    
    branch_stack = db.relationship(
        'Branch',
        backref='part',
        order_by='Branch.index',
        collection_class=ordering_list('index'),
        foreign_keys='Branch._part_id'
    )
        
    current_branch = db.relationship(
        'Branch',
        uselist=False,
        foreign_keys='Branch._part_head_id'
    )
    
    @property
    def current_page(self):
        return self.current_branch.current_page
        
    @property
    def pages(self):
        return [p for b in self.branch_stack for p in b.pages]

    embedded = db.relationship(
        'Embedded',
        backref='part',
        order_by='Embedded.index',
        collection_class=ordering_list('index'),
    )
        
    @property
    def data_elements(self):
        elements = [e for e in self.embedded]
        elements += [e for b in self.branch_stack for e in b.data_elements]
        elements.sort(key=lambda e: e.id)
        return elements
    
    _viewing_pages = db.relationship(
        'ViewingPage',
        backref='part',
        order_by='ViewingPage.index',
        collection_class=ordering_list('index')
    )

    _previous_status = db.Column(db.String(16))
    cached_data = db.Column(MutableDictType, default={})
    _completed = db.Column(db.Boolean, default=False)
    end_time = db.Column(db.DateTime)
    g = db.Column(MutableDictType)
    meta = db.Column(MutableDictType)
    start_time = db.Column(db.DateTime)
    _time_expired = db.Column(db.Boolean, default=False)
    updated = db.Column(db.Boolean, default=True)
    
    @property
    def completed(self):
        return self._completed
        
    @completed.setter
    @_send_data
    def completed(self, completed):
        self._completed = completed
        
    @property
    def time_expired(self):
        return self._time_expired
        
    @time_expired.setter
    @_send_data
    def time_expired(self, time_expired):
        self._time_expired = time_expired
    
    @property
    def status(self):
        if self.completed:
            return 'Completed'
        if self.time_expired:
            return 'TimedOut'
        return 'InProgress'
    
    def __init__(self, **kwargs):
        self._key = key()
        self.end_time = self.start_time = datetime.utcnow()
        super().__init__(**kwargs)

        ds = DataStore.query.first()
        ds.meta.append(self.meta)
        if not settings.get('collect_IP'):
            self.meta.pop('IPv4', None)
        ds.update_status(self)

    def _init_tree(self, gen_root):
        """
        Initialize a new tree.
        
        Parameters
        ----------
        gen_root : callable
            Generates the root branch of the tree.

        Returns
        -------
        self
        """
        self._router = Router(gen_root)
        self.current_branch = root = gen_root()
        self.branch_stack.append(root)
        root.current_page = root.start_page
        if not self.current_branch.pages:
            self._router.navigator.forward_recurse()
        return self

    def back(self, back_to=None):
        """
        Navigate back for debugging purposes.

        Parameters
        ----------
        back_to : hemlock.Page or None, default=None
            Navigate back to this page; if `None`, navigate back one page.

        Returns
        -------
        self : hemlock.Participant
        """
        self._router.navigator.back(back_to)
        return self

    def forward(self, forward_to=None):
        """
        Navigate forward for debugging purposes.

        Parameters
        ----------
        forward_to : hemlock.Page or None, default=None
            Navigate forward to this page; if `None`, navigate forward one 
            page.

        Returns
        -------
        self : hemlock.Participant
        """
        self._router.navigator.forward(forward_to)
        return self

    def gen_test_participant(gen_root=None):
        """
        Generate a test participant for debugging purposes.

        Parameters
        ----------
        gen_root : callable or None, default=None
            Function to generate the root branch of the participant's tree.
            This should return a `hemlock.Branch`.

        Returns
        -------
        part : hemlock.Participant
        """
        part = Participant(meta={})
        login_user(part)
        if gen_root:
            part._init_tree(gen_root)
        return part

    def get_data(self):
        """
        Returns
        -------
        df : hemlock.models.private.DataFrame
            Data associated with the participant.

        Examples
        --------

        Notes
        -----
        Data elements are added to the dataframe in the order in which they 
        were created (i.e. by id). This is not necessarily the order in which 
        they appeared to the Participant.
        """
        self._set_order()
        elements = self.data_elements
        df = DataFrame()
        df.add(data=self.get_meta(), rows=-1)
        [df.add(data=e._pack_data(), rows=e.data_rows) for e in elements]
        df.pad()
        self.cached_data = df
        return df
    
    def _set_order(self):
        """Set the order for all data elements
        
        An element's order is the order in which it appeared to the 
        Participant relative to other elements of the same variable. These 
        functions walk through the survey and set the element order.
        
        The participant's embedded data are set first, followed by its
        branches' data. A branch's embedded data are set first, followed by
        its page's data. A page's embedded data are set first, followed by its
        timer, followed by its quesitons.
        """
        var_count = {}
        [self._set_order_element(e, var_count) for e in self.embedded]
        if self.branch_stack:
            self._set_order_branch(self.branch_stack[0], var_count)
    
    def _set_order_branch(self, branch, var_count):
        """Set the order for Questions belonging to a given Branch"""
        [self._set_order_element(e, var_count) for e in branch.embedded]
        [self._set_order_page(p, var_count) for p in branch.pages]
        if branch.next_branch in self.branch_stack:
            self._set_order_branch(branch.next_branch, var_count)
    
    def _set_order_page(self, page, var_count):
        """Set the order for Questions belonging to a given Page"""
        elements = page.data_elements
        [self._set_order_element(e, var_count) for e in elements]
        if page.next_branch in self.branch_stack:
            self._set_order_branch(page.next_branch, var_count)
        
    def _set_order_element(self, element, var_count):
        """Set the order for a given data element"""
        var = element.var
        if var is None:
            return
        if var not in var_count:
            var_count[var] = 0
        element.order = var_count[var]
        var_count[var] += 1
    
    def get_meta(self):
        """
        This is where it gets meta.

        Returns
        -------
        meta : dict
            Participant's metadata, including the ID, end time, start time, 
            and current status.
        """
        meta = self.meta.copy()
        meta.update({
            'ID': self.id,
            'EndTime': self.end_time,
            'StartTime': self.start_time,
            'Status': self.status
        })
        return meta

    def view_nav(self):
        """
        View participant's branch stack.

        Returns
        -------
        self : hemlock.Participant
        """
        self.branch_stack[0].view_nav()
        print('\n C = current page \n T = terminal page\n')
        return self


"""# Navigation bar

Examples
--------
```python
from hemlock import Page, push_app_context
from hemlock.tools import Navbar, Navitem, Navitemdd, Dropdownitem

app = push_app_context()

url_root = 'https://dsbowen.github.io/'

navbar = Navbar(
\    'Hemlock', 
\    [
\        ('Application', href=url_root+'app'),
\        (
\            'Tools', 
\            [
\                ('Language', href=url_root+'tools/lang'),
\                ('Navbar', href=url_root+'tools/navbar')
\           ]
\        )
\    ], 
\    href=url_root+'hemlock'
)

Page(navbar=navbar.render()).preview()
```
"""

from .random import key
from .statics import format_attrs

from convert_list import ConvertList
from flask import request

import os
from copy import copy, deepcopy

DIR = os.path.dirname(os.path.realpath(__file__))

class NavBase():
    """
    All navigation models inherit from this base.

    Parameters and attributes
    -------------------------
    label : str
        Label for the navigation element.

    template : str
        Path to template file or string. This is *not* a Jinja template, as 
        you may wish to generate html for statics outside the application 
        context.

    href : str, default=''
        Hyperref to which the navigation element links.

    a_attrs : dict, default={}
        Dictionary of HTML attributes for the navigation element's `<a>` tag.
    """
    @property
    def href(self):
        return self.a_attrs.get('href')

    @href.setter
    def href(self, val):
        self.a_attrs['href'] = val

    def __init__(self, label, template, href='', a_class=[], a_attrs={}):
        self.id = 'navbar-'+key(10)
        self.label = label
        self.template = (
            open(template).read() if os.path.exists(template)
            else template # assume template is string
        )
        self.a_class = a_class.copy()
        self.a_attrs = deepcopy(a_attrs)
        self.href = href

    def render(self):
        """
        Returns
        -------
        rendered : str
        """
        return self.template.format(self=self, a_attrs=self._render_attrs())

    def _render_attrs(self):
        a_class = self.a_class.copy()
        if self.href == str(request.url_rule):
            a_class.append('active')
        return format_attrs(**{'class': a_class}, **self.a_attrs)


class NavitemList(ConvertList):
    @classmethod
    def convert(cls, item):
        if isinstance(item, (Navitem, Navitemdd)):
            return item
        # item is tuple (label, href) if Navitem 
        # or (label, [dropdownitems]) if Navitemdd
        if isinstance(item[1], str):
            return Navitem(label=item[0], href=item[1])
        if isinstance(item[1], list):
            return Navitemdd(label=item[0], dropdownitems=item[1])
        raise ValueError('Incorrect value for navitem')


class Navbar(NavBase):
    """
    Parameters
    ----------
    label : str, default=''

    navitems : list of hemlock.tools.Navitem and hemlock.tools.Navitemdd
        Navigation items associated with the navbar.

    href : str, default=''

    template : str, default='directory/navbar.html'
        By default, this is a file stored in the directory of the current 
        file.

    a_attrs : dict

    navbar_attrs : dict
        HTML attributes for the `<nav>` tag.
    """
    @property
    def navitems(self):
        return self._navitems

    @navitems.setter
    def navitems(self, items):
        self._navitems = NavitemList(items)

    def __init__(
            self, label='', navitems=[], href='#',
            template=os.path.join(DIR, 'navbar.html'), 
            a_class=['navbar-brand'], a_attrs={},
            navbar_attrs={
                'class': [
                    'navbar', 
                    'navbar-expand-lg', 
                    'navbar-light', 
                    'bg-light', 
                    'fixed-top'
                ]
            }
        ):
        super().__init__(label, template, href, a_class, a_attrs)
        self.navitems = navitems
        self.navbar_attrs = deepcopy(navbar_attrs)

    def render(self):
        """
        Returns
        -------
        rendered : str
            Rendered navbar HTML.
        """
        return self.template.format(
            self=self,
            navbar_attrs=format_attrs(**self.navbar_attrs),
            a_attrs=self._render_attrs(),
            navitems='\n'.join([item.render() for item in self.navitems])
        )


class Navitem(NavBase):
    """
    Navigation item *without* dropdown items.

    Parameters
    ----------
    label : str, default=''

    href : str, default=''

    template : str, default='directory/navitem.html'
        By default, this is a file stored in the directory of the current 
        file.

    a_attrs : dict
    
    navitem_attrs : dict
        HTML attributes for the `navitem` div.
    """
    def __init__(
            self, label='', href='',
            template=os.path.join(DIR, 'navitem.html'), 
            a_class=['nav-item', 'nav-link'], a_attrs={}
        ):
        super().__init__(label, template, href, a_class, a_attrs)


class DropdownitemList(ConvertList):
    @classmethod
    def convert(cls, item):
        if isinstance(item, Dropdownitem):
            return item
        # item is tuple (label, href)
        if isinstance(item, tuple):
            return Dropdownitem(label=item[0], href=item[1])
        raise ValueError('Incorrect value for dropdown item')


class Navitemdd(NavBase):
    """
    Navigation item *with* dropdown items.

    Parameters
    ----------
    label : str, default=''

    dropdownitems : list of hemlock.tools.Dropdownitem
    
    template : str, default='directory/navitemdd.html'
        By default, this is a file stored in the directory of the current 
        file.

    a_attrs : dict

    navitem_attrs : dict
        HTML attributes for the `navitem` div.
    """
    @property
    def dropdownitems(self):
        return self._dropdownitems

    @dropdownitems.setter
    def dropdownitems(self, items):
        self._dropdownitems = DropdownitemList(items)

    def __init__(
            self, label='', dropdownitems=[], 
            template=os.path.join(DIR, 'navitemdd.html'),
            a_class=['nav-item', 'nav-link', 'dropdown', 'dropdown-toggle'],
            a_attrs={
                'role': 'button',
                'data-toggle': 'dropdown',
                'aria-haspopup': 'true',
                'aria-expanded': 'false'
            },
            navitem_attrs={'class': ['nav-item', 'dropdown']}
        ):
        # note: href should not be passed
        super().__init__(label, template, a_class=a_class, a_attrs=a_attrs)
        self.dropdownitems = dropdownitems

    def render(self):
        """
        Returns
        -------
        rendered : str
        """
        return self.template.format(
            self=self,
            a_attrs=self._render_attrs(),
            items='\n'.join([item.render() for item in self.dropdownitems])
        )


class Dropdownitem(NavBase):
    """
    Parameters
    ----------
    label : str, default=''

    href : str, default=''

    template : str, default='directory/dropdownitem.html'
        By default, this is a file stored in the directory of the current 
        file.

    a_attrs : dict
    """
    def __init__(
            self, label='', href='',
            template=os.path.join(DIR, 'dropdownitem.html'),
            a_class=['dropdown-item'], a_attrs={}
        ):
        super().__init__(label, template, href, a_class, a_attrs)


"""# Choices and Options

The difference between `hemlock.Choice` and `hemlock.Option` is the former are 
for `hemlock.Check` questions, while latter are for `hemlock.Select` questions.

The use of choice and option models is not due to any deep functional 
difference between them, but reflects the underlying html.
"""

from ..app import db, settings
from ..models import Base
from ..tools import format_attrs, key
from .bases import InputBase

from convert_list import ConvertList
from flask import render_template
from sqlalchemy.types import JSON, TypeDecorator
from sqlalchemy_mutable import MutableList
import json
import time


class ChoiceBase(InputBase, Base):
    """
    Base class for choices.

    Parameters
    ----------
    label : str or None, default=None
        Choice label.

    template : str
        Jinja template for the choice html.

    value : default=None
        Value of the choice if selected. e.g. a choice with label `'Yes'` 
        might have a value of `1`. If `None`, the `label` is used. For a 
        question where only one choice can be selected, this is the value of 
        the question's data if this choice is selected. For a question where 
        multiple choices may be selected, data are one-hot encoded; the value 
        is the suffix of the column name associated with the indicator 
        variable that this choice was selected.

    name : default=None
        Name associated with this choice in the dataframe. If `None`, the 
        `label` is used.

    Attributes
    ----------
    key : str
        Randomly generated from ascii letters and digits.

    label : str
        Set from the `label` parameter.

    value : 
        Set from the `value` parameter.

    name :
        Set from the `name` parameter.

    Notes
    -----
    If passing `value` and `name` to contructor, these must be passed as 
    keyword arguments
    """
    def __init__(self, label=None, **kwargs):
        self.key = key(10)
        self.set_all(label)
        super().__init__(**kwargs)

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, self.value.__repr__())

    def is_default(self, question):
        """
        Parameters
        ----------
        question : hemlock.Question
            The question to which this choice belongs.

        Returns
        -------
        is_default : bool
            Indicate that this choice is (one of) its question's default 
            choice(s).

        Notes
        -----
        The question's default choice(s) is the question's `response` if the 
        participant responded to the question, or the question's `default` if
        the participant has not yet responded to the question.
        """
        default = (
            question.response if question.has_responded else question.default
        )
        if default in (None, []):
            return False
        return (
            self.value in default if isinstance(default, list)
            else self.value == default
        )

    def set_all(self, val):
        """
        Set the choice's label, name, and value.

        Parameters
        ----------
        val : 
            Value to which the choice's label, name, and value should be set.

        Returns
        -------
        self : hemlock.Choice
        """
        self.label = self.name = self.value = val
        return self

    def dump(self):
        return json.dumps(self.__dict__)

    @classmethod
    def load(cls, state_dict):
        return cls(**json.loads(state_dict))


class ChoiceListTypeBase(TypeDecorator):
    impl = JSON
    choice_cls = None # must be implemented by ChoiceListType

    def process_bind_param(self, choices, dialect):
        return [choice.dump() for choice in choices]

    def process_result_value(self, state_dicts, dialect):
        return [self.choice_cls.load(state) for state in state_dicts]


class ChoiceListBase(ConvertList, MutableList):
    """
    Converts items in a list of choices to the choice class.
    """
    choice_cls = None # must be implemented by ChoiceList

    @classmethod
    def convert(cls, item):
        if isinstance(item, cls.choice_cls):
            # item is already the choice class
            return item
        if isinstance(item, dict):
            # item is a dict of keyword arguments for the choice class
            return cls.choice_cls(**item)
        if isinstance(item, (tuple, list)):
            if len(item) == 2:
                # item is (label, value) tuple
                return cls.choice_cls(label=item[0], value=item[1])
            if len(item) == 3:
                # item is (label, value, name) tuple
                return cls.choice_cls(
                    label=item[0], value=item[1], name=item[2]
                )
        # item is label, value, name
        return cls.choice_cls(label=str(item), value=item, name=str(item))


settings['Choice'] = dict(
    div_class=['custom-control'], div_attrs={},
    input_attrs={'class': ['custom-control-input']},
    label_attrs={'class': ['custom-control-label', 'w-100', 'choice']}
)

class Choice(ChoiceBase):
    """
    Choices are displayed as part of their question (usually 
    `hemlock.Check`). Inherits from `hemlock.ChoiceBase`.

    Parameters
    ----------
    label : str or None, default=None
        Choice label.

    template : str, default='hemlock/choice.html'
        File name of the choice template.

    \*\*kwargs :
        Set the choice's value and name using keyword arguments.
    """    
    def __init__(self, label=None, template='hemlock/choice.html', **kwargs):
        super().__init__(label, template=template, **kwargs)

    def click(self, driver, if_selected=None):
        """
        Use a selenium webdriver to click on this choice.

        Parameters
        ----------
        driver : selenium.webdriver.chrome.webdriver.WebDriver
            The selenium webdriver that clicks this choice. Does not have to 
            be chromedriver.

        if_selected : bool or None, default=None
            Indicates that the choice will be clicked only if it is already 
            selected. If `False` the choice will be clicked only if it is not 
            already selected. If `None` the choice will be clicked whether or 
            not it is selected.

        Returns
        -------
        self
        """
        inpt = driver.find_element_by_css_selector('#'+self.key)
        if (
            if_selected is None
            or if_selected == bool(inpt.get_attribute('checked'))
        ):
            selector = 'label[for={}]'.format(self.key)
            driver.find_element_by_css_selector(selector).click()
        return self

    def is_displayed(self, driver):
        """
        Parameters
        ----------
        driver : selenium.webdriver.chrome.webdriver.WebDriver
            The selenium webdriver that clicks this choice. Does not have to 
            be chromedriver.

        Returns
        -------
        is_displayed : bool
            Indicates that this choice is visible in the browser.
        """
        selector = 'label[for={}]'.format(self.key)
        label = driver.find_element_by_css_selector(selector)
        return label.is_displayed()

    def _render(self, question, idx):
        return render_template(
            self.template, c=self, q=question, idx=idx,
            div_attrs=format_attrs(**self.div_attrs),
            input_attrs=format_attrs(**self.input_attrs),
            label_attrs=format_attrs(**self.label_attrs) 
        )

class ChoiceListType(ChoiceListTypeBase):
    choice_cls = Choice

class ChoiceList(ChoiceListBase):
    choice_cls = Choice

ChoiceList.associate_with(ChoiceListType)


settings['Option'] = dict(input_attrs={})

class Option(ChoiceBase):
    """
    Options are displayed as part of their question (usually 
    `hemlock.Select`). Inherits from `hemlock.ChoiceBase`. Its functionality
    is similar to `hemlock.Choice`, but for `Select` questions instead of
    `Check` questions.

    Parameters
    ----------
    label : str or None, default=None
        Choice label.

    template : str, default='hemlock/option.html'
        File name of the option template.

    \*\*kwargs :
        Set the choice's value and name using keyword arguments.
    """
    _input_attr_names = ['class', 'disabled']

    def __init__(self, label=None, template='hemlock/option.html', **kwargs):
        super().__init__(label, template=template, **kwargs)

    def click(self, driver, if_selected=None):
        """
        Use a selenium webdriver to click on this choice.

        Parameters
        ----------
        driver : selenium.webdriver.chrome.webdriver.WebDriver
            The selenium webdriver that clicks this choice. Does not have to 
            be chromedriver.

        if_selected : bool or None, default=None
            Indicates that the choice will be clicked only if it is already 
            selected. If `False` the choice will be clicked only if it is not 
            already selected. If `None` the choice will be clicked whether or 
            not it is selected.

        Returns
        -------
        self
        """
        option = driver.find_element_by_css_selector('#'+self.key)
        if (
            if_selected is None 
            or if_selected == bool(option.get_attribute('selected'))
        ):
            option.click()
        return self

    def is_displayed(self, driver):
        """
        Parameters
        ----------
        driver : selenium.webdriver.chrome.webdriver.WebDriver
            The selenium webdriver that clicks this choice. Does not have to 
            be chromedriver.

        Returns
        -------
        is_displayed : bool
            Indicates that this choice is visible in the browser.
        """
        option = driver.find_element_by_css_selector('#'+self.key)
        return option.is_displayed()

    def _render(self, question, idx):
        return render_template(
            self.template, opt=self, q=question, idx=idx, 
            option_attrs=format_attrs(**self.input_attrs)
        )


class OptionListType(ChoiceListTypeBase):
    choice_cls = Option

class OptionList(ChoiceListBase):
    choice_cls = Option

OptionList.associate_with(OptionListType)


"""# Range slider"""

from ..app import db, settings
from ..functions.debug import drag_range, send_keys
from ..models import Question
from .bases import InputBase

from flask import render_template

settings['Range'] = {
    'debug': drag_range,
    'type': 'range', 
    'class': ['custom-range'], 
    'min': 0, 'max': 100, 'step': 1
}


class Range(InputBase, Question):
    """
    Range sliders can be dragged between minimum and maximum values in step 
    increments.

    Inherits from [`hemlock.InputBase`](bases.md) and 
    [`hemlock.Question`](../models/question.md).

    Parameters
    ----------
    label : str or None, default=None
        Range label.

    template : str, default='hemlock/range.html'
        Template for the range body.

    Notes
    -----
    Ranges have a default javascript which displays the value of the range 
    slider to participants. This will be appended to any `js` and `extra_js`
    arguments passed to the constructor.

    Examples
    --------
    ```python
    from hemlock import Range, Page, push_app_context

    app = push_app_context()

    Page(Range('<p>This is a range slider.</p>')).preview()
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'range'}

    def __init__(self, label=None, template='hemlock/range.html', **kwargs):
        super().__init__(label=label, template=template, **kwargs)
        self.js.append(render_template('hemlock/range.js', q=self))


settings['RangeInput'] = {
    'debug': send_keys,
    'width': '5em',
    'class': ['form-control'],
    'type': 'number', 'min': 0, 'max': 100, 'step': 1
}


class RangeInput(InputBase, Question):
    """
    Range slider with an input field.

    Inherits from [`hemlock.InputBase`](bases.md) and 
    [`hemlock.Question`](../models/question.md).

    Parameters
    ----------
    label : str or None, default=None
        Range label.

    template : str, default='hemlock/rangeinput.html'
        Template for the range body.

    width : str, default='5em'
        Width of the input field.

    Notes
    -----
    Ranges have a default javascript which displays the value of the range 
    slider to participants. This will be appended to any `js` and `extra_js`
    arguments passed to the constructor.

    Examples
    --------
    ```python
    from hemlock import RangeInput, Page, push_app_context

    app = push_app_context()

    Page(RangeInput('<p>This is a range slider.</p>')).preview()
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'rangeinput'}

    width = db.Column(db.String)

    def __init__(
            self, label=None, template='hemlock/rangeinput.html', **kwargs
        ):
        super().__init__(label=label, template=template, **kwargs)
        self.js.append(render_template('hemlock/rangeinput.js', q=self))


"""Public database models, question polymorphs, and tools"""

from .app import create_app, db, push_app_context, settings
from .models import (
    Branch, Embedded, Timer, Compile, Debug, Validate, Submit, Navigate, 
    Page, Participant, Question, ChoiceQuestion, Worker
)
from .qpolymorphs import (
    Binary, Blank, Check, Choice, Dashboard, Download, File, Input, Label, 
    Range, RangeInput, Select, Option, Textarea
)
from .routes import route
from .functions import compile, debug, submit, validate
from . import tools
from . import routes


"""Public database models"""

from .bases import Base, Data
from .branch import Branch
from .embedded import Embedded, Timer
from .functions import Compile, Debug, Validate, Submit, Navigate
from .page import Page
from .participant import Participant
from .question import Question, ChoiceQuestion
from .worker import Worker


"""# Comprehension check"""

ERROR_MSG = """
Your response was incorrect.<br/>
Please reread the instructions before continuing.
"""

def comprehension_check(instructions, checks, attempts=None):
    """
    A comprehension check consists of 'instruction' pages followed by 'check' 
    pages. The data of all questions in a check page must evaluate to `True` 
    to pass the check. When a participant fails a check, he is brought back to 
    the first instructions page. 
    
    Participants only have to pass each check once. For example, suppose there 
    are two checks, A and B. The participant passes check A but fails check B. 
    He is brought back to the first page of the instructions. After rereading 
    the instructions, he is brought directly to check B, skipping check A.

    Parameters
    ----------
    instructions : hemlock.Page or list of hemlock.Page
        Instruction page(s).

    checks : hemlock.Page or list of hemlock.Page
        Check page(s).

    attempts : int or None, default=None
        Number of attempts allotted. Participants are allowed to proceed with 
        the survey after exceeding the maximum number of attempts. If `None`, 
        participants must pass the comprehension check before continuing the 
        survey.

    Returns
    -------
    pages : list of hemlock.Page
        List of instructions pages + check pages.

    Notes
    -----
    This function adds a `hemlock.Submit` function to each check page. This 
    must be the last submit function of each check page. 

    Examples
    --------
    We have two files in our root directory. In `survey.py`:

    ```python
    from hemlock import Branch, Page, Label, Input, Submit as S, route
    from hemlock.tools import comprehension_check

    @route('/survey')
    def start():
    \    return Branch(
    \        *comprehension_check(
    \            instructions=Page(
    \                Label('<p>Here are some instructions.</p>')
    \            ),
    \            checks=Page(
    \                Input(
    \                    '<p>Enter "hello world" or you... shall not... PASS!</p>',
    \                    submit=S.match('hello world')
    \                )
    \            )
    \        ),
    \        Page(
    \            Label('<p>You passed the comprehension check!</p>'),
    \            terminal=True
    \        )
    \    )
    ```

    In `app.py`:

    ```python
    import survey

    from hemlock import create_app

    app = create_app()

    if __name__ == '__main__':
    \    from hemlock.app import socketio
    \    socketio.run(app, debug=True)
    ```

    Run the app with:

    ```
    $ python app.py # or python3 app.py
    ```

    Open your browser to <http://localhost:5000/>.
    """
    assert instructions and checks, (
        '`instructions` and `checks` must be non-empty lists of hemlock.Pages'
    )
    from ..models import Submit
    if not isinstance(instructions, list):
        instructions = [instructions]
    if not isinstance(checks, list):
        checks = [checks]
    for check in checks:
        check.back_to = instructions[0]
        check.submit.append(Submit(
            _verify_data, 
            instructions[-1], 
            curr_attempt=1, 
            attempts=attempts
        ))
    return instructions + checks

def _verify_data(check, last_instr_page, curr_attempt, attempts):
    """
    Verify that the data on the comprehension page is correct.

    Parameters
    ----------
    check : hemlock.Page
        Check page.

    last_instr_page : hemlock.Page
        Last instructions page.

    curr_attempt : int
        Current attempt number.

    attempts : int or None
        Maximum number of attempts.
    """
    if (
        all(q.data for q in check.questions) 
        or (attempts is not None and curr_attempt >= attempts)
    ):
        if check != check.branch.pages[-1]:
            # this check does not have to be repeated
            last_instr_page.forward_to = check.branch.pages[check.index+1]
            pass
        return
    check.back_to.error = ERROR_MSG
    check.direction_from = 'back'
    check.submit[-1].kwargs['curr_attempt'] = curr_attempt + 1


"""# Function column types

All of these classes inherit from 
[`sqlalchemy_mutable.partial`](https://dsbowen.github.io/sqlalchemy-mutable).
"""

from sqlalchemy_mutable import partial

from random import random


class Compile(partial):
    """
    Helps compile a page or question html before it is rendered and displayed 
    to a participant.

    Examples
    --------
    ```python
    from hemlock import Compile as C, Input, Label, Page, push_app_context

    app = push_app_context()

    @C.register
    def greet(greet_q, name_q):
    \    greet_q.label = '<p>Hello {}!</p>'.format(name_q.response)

    name_q = Input("<p>What's your name?</p>")
    p = Page(Label(compile=C.greet(name_q)))
    name_q.response = 'World'
    p._compile().preview()
    ```
    """


class Debug(partial):
    """
    Run to help debug the survey.

    Examples
    --------
    ```python
    from hemlock import Debug as D, Input, Page, push_app_context
    from hemlock.tools import chromedriver

    app = push_app_context()

    driver = chromedriver()

    @D.register
    def greet(driver, greet_q):
    \    inpt = greet_q.input_from_driver(driver)
    \    inpt.clear()
    \    inpt.send_keys('Hello World!')

    p = Page(Input('<p>Enter a greeting.</p>', debug=D.greet()))
    p.debug.pop(-1) # so the page won't navigate
    p.preview(driver)._debug(driver)
    ```
    """
    def __init__(self, func, *args, p_exec=1., **kwargs):
        super().__init__(func, *args, **kwargs)
        self.p_exec = p_exec

    def __call__(self, *args, **kwargs):
        if random() < self.p_exec:
            return super().__call__(*args, **kwargs)


class Validate(partial):
    """
    Validates a participant's response.

    Attributes
    ----------
    error_msg : str or None
        If the validate function returns an error message, `error_msg` is 
        returned instead of the output of the validate function.
        You can set this by passing in an `error_msg` keyword argument to the
        constructor.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    @V.register
    def match(inpt, pattern):
    \    if inpt.response != pattern:
    \        return '<p>You entered "{}", not "{}"</p>'.format(inpt.response, pattern)

    pattern = 'hello world'
    inpt = Input(validate=V.match(pattern))
    inpt.response = 'goodbye moon'
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    You entered "goodbye moon", not "hello world"
    ```
    """
    def __init__(self, func, *args, error_msg=None, **kwargs):
        super().__init__(func, *args, **kwargs)
        self.error_msg = error_msg

    def __call__(self, *args, **kwargs):
        error_msg = super().__call__(*args, **kwargs)
        if error_msg:
            return self.error_msg or error_msg

            
class Submit(partial):
    """
    Runs after a participant has successfully submitted a page.

    Examples
    --------
    ```python
    from hemlock import Input, Submit as S, push_app_context

    app = push_app_context()

    @S.register
    def get_initials(name_q):
    \    names = name_q.response.split()
    \    name_q.data = '.'.join([name[0] for name in names]) + '.'

    inpt = Input("<p>What's your name?</p>", submit=S.get_initials())
    inpt.response = 'Andrew Yang'
    inpt._submit().data
    ```

    Out:

    ```
    A.Y.
    ```
    """


class Navigate(partial):
    """
    Creates a new branch to which the participant will navigate.

    Examples
    --------
    ```python
    from hemlock import Branch, Navigate as N, Page, Participant, push_app_context

    def start():
    \    return Branch(Page(), navigate=N.end())

    @N.register
    def end(start_branch):
    \    return Branch(Page(terminal=True))

    app = push_app_context()

    part = Participant.gen_test_participant(start)
    part.view_nav()
    ```

    Out:

    ```
    <Branch 1>
    <Page 1> C 

    C = current page 
    T = terminal page
    ```

    In:

    ```python
    part.forward().view_nav()
    ```

    Out:

    ```
    <Branch 1>
    <Page 1>  
    \    <Branch 2>
    \    <Page 2> C T

    C = current page 
    T = terminal page
    ```
    """


"""Viewing page"""

from ...app import db

from bs4 import BeautifulSoup
from flask import request

import os
from tempfile import NamedTemporaryFile


class ViewingPage(db.Model):
    """
    Stores HTML snapshot of each page for each participant. These can be 
    accessed in the researcher dashboard.

    Parameters
    ----------
    html : str
        HTML of the page.

    first_presentation : bool
        Indicates that this was the first time this page was presented to the 
        participant.

    Attributes
    ----------
    first_presentation : bool
        Set from the `first_presentation` parameter.

    html : str
        Set from the `html` parameter.

    index : str
        Order in which the participant saw this page.

    url_root : str
        URL root for viewing this page. This will be used to get absolute 
        paths to statics.

    Relationships
    -------------
    part : hemlock.Participant
        Set from the `part` parameter.
    """
    id = db.Column(db.Integer, primary_key=True)
    downloaded = db.Column(db.Boolean, default=False)
    part_id = db.Column(db.Integer, db.ForeignKey('participant.id'))
    first_presentation = db.Column(db.Boolean)
    html = db.Column(db.String)
    index = db.Column(db.Integer)
    url_root = db.Column(db.String)
    
    def __init__(self, html, first_presentation=True):
        self.html = html
        self.first_presentation = first_presentation
        self.url_root = request.url_root

    def mkstmp(self):
        """
        Returns
        -------
        path : str
            Path to temporary html file.
        """
        def convert_rel_paths():
            """
            Convert stylesheets and scripts from relative to absolute paths and 
            remove banner.
            """
            soup = BeautifulSoup(self.html, 'html.parser')
            convert_url_attr(soup, 'href')
            convert_url_attr(soup, 'src')
            banner = soup.select_one('#banner')
            if banner is not None:
                banner.extract()
            self.html = str(soup)

        def convert_url_attr(soup, url_attr):
            elements = soup.select('[{}]'.format(url_attr))
            for e in elements:
                url = e.attrs.get(url_attr)
                if url is not None and url.startswith('/'):
                    e.attrs[url_attr] = self.url_root + url

        if not self.downloaded:
            convert_rel_paths()
            self.downloaded = True # cache result
        with NamedTemporaryFile('w', suffix='.html', delete=False) as f:
            f.write(self.html)
            uri = f.name
        return uri


"""# Titration tool

Titration is used primarily to elicit willingness to pay or risk preferences.
"""

from ..app import db


def titrate(
        gen_titrate_q, distribution, var, data_rows=1, 
        tol=None, max_pages=None, **kwargs
    ):
    """
    Parameters
    ----------
    gen_titrate_q : callable
        Callable which takes the titration value and returns a 
        `hemlock.Question`. If the question's `data` evaluates to `True`, the
        next iteration will increase the titration value. If the question's 
        `data` evaluates to `False`, the next iteration will decrease the
        titration value.

    distribution : callable
        Callable with a `ppf` method which takes a quantile `q` and returns
        a value (e.g. `float`). I recommend `scipy.stats` distributions.

    var : str
        Name of the variable in which to store the titrated value.

    data_rows : int, default=1
        Number of data rows for the titration variable.

    tol : float or None, default=None
        Titration will stop when the titrated value reaches this precision. If
        `None`, titration stops after a certain number of questions.

    max_pages : int or None, default=None
        Titration will stop after this number of pages. If `None`, titration
        stops after a precision level is reached.

    \*\*kwargs :
        Keyword arguments are passed to the titration page constructor.

    Returns
    -------
    FIrst titration page : hemlock.Page
    
    Examples
    --------
    This example demonstrates how to use titration to more efficiently and
    precisely implement the 
    <a href="http://www.planchet.net/EXT/ISFA/1226.nsf/9c8e3fd4d8874d60c1257052003eced6/9ba2376a9b20919cc125784b00355f88/$FILE/HOLT.pdf" target="_blank">Holt-Laury risk elicitation method</a>. 
    The titration value is the probability at which the participant is 
    indifferent between two lotteries.

    We then compute the coefficient of relative risk aversion assuming CRRA 
    utility.


    ```python
    import numpy as np
    from hemlock import Branch, Page, Binary, Label, Validate as V, route
    from hemlock.tools import titrate
    from scipy.optimize import minimize
    from scipy.stats import uniform

    import math

    def gen_titrate_q(p_larger):
    \    # p_larger is the probability of winning the larger amount of money
    \    p = round(p_larger)
    \    return Binary(
    \        '''
    \        Which lottery would you rather have?
    \        ''',
    \        [
    \            '{}/100 of $2.00, {}/100 of $1.60'.format(p, 100-p),
    \            '{}/100 of $3.85, {}/100 of $0.10'.format(p, 100-p)
    \        ],
    \        inline=False, validate=V.require()
    \    )

    @route('/survey')
    def start():
    \    return Branch(
    \        titrate(gen_titrate_q, uniform(0, 100), var='p', tol=5, back=True),
    \        Page(
    \            Label(compile=disp_risk_aversion),
    \            back=True, terminal=True,
    \        )
    \    )

    def disp_risk_aversion(label):
    \    def func(r):
    \        def u(x):
    \            # CRRA utility function
    \            return x**(1-r) if abs(1-r) > .01 else math.log(x)

    \        return (p*u(2)+(1-p)*u(1.6)-p*u(3.85)-(1-p)*u(.1))**2

    \    p = label.part.g['p'] / 100. # indifference point between 0 and 1
    \    # coefficient of relative risk aversion if optimization starts from -3
    \    res_lower = minimize(func, x0=np.array([-3])).x
    \    # coefficient of relative risk aversion if optimization starts from 2
    \    res_upper = minimize(func, x0=np.array([2])).x
    \    # you can trivially minimize the function by setting r == 1 (approximately)
    \    # this sets u(x) = 0 (approx.) for all x
    \    # so choose the r farther from 1
    \    r = res_lower if abs(res_lower-1)>abs(res_upper-1) else res_upper
    \    label.label = '''
    \        Indifference point is {:.0f}/100. Coefficient of relative risk aversion is {:.2f}
    \        '''.format(100.*p, float(r))
    ```
    """
    assert tol is not None or max_pages is not None
    n_pages_left = None if max_pages is None else max_pages - 1
    # initial titration value is the median of the distribution
    x = distribution.ppf(.5)
    return _titrate(
        gen_titrate_q, x, distribution, .5, .25, var, data_rows, tol, 
        n_pages_left, **kwargs
    )

def _titrate(
        gen_titrate_q, x, distribution, q, step, var, data_rows, tol, 
        n_pages_left, **kwargs
    ):
    from ..models import Page, Submit as S

    page =  Page(gen_titrate_q(x), **kwargs)
    # submit function computes the next titration value
    page.submit.append(
        S(
            _gen_next_titrate_page, gen_titrate_q, x, distribution, q, step, 
            var, data_rows, tol, n_pages_left, **kwargs
        )
    )
    return page

def _gen_next_titrate_page(
        page, gen_titrate_q, prev_x, distribution, q, step, var, data_rows, 
        tol, n_pages_left, **kwargs
    ):
    def remove_titration_pages():
        # remove pages in front of this page which are part of the titration
        # e.g. after navigating back during titration, you will have stale
        # titration pages
        while (
            page != page.branch.pages[-1] 
            and isinstance(page.branch.pages[page.index+1].g, dict)
            and page.branch.pages[page.index+1].g.get('Titration')
        ):
            stale_titration_page = page.branch.pages.pop(page.index+1)
            db.session.delete(stale_titration_page)

    def record_x(x):
        # record the titration value as embedded data and in the participant's
        # `g`
        from ..models import Embedded

        x = float(x)
        page.part.embedded.append(Embedded(var, x, data_rows=data_rows))
        page.part.g[var] = x
        
    remove_titration_pages()
    # step the titration value
    q += step if page.questions[-1].data else -step
    x = distribution.ppf(q)
    if (
        n_pages_left == 0 
        or (tol is not None and abs(x - prev_x) < tol)
    ):
        return record_x(x)
    step /= 2.
    n_pages_left = None if n_pages_left is None else n_pages_left - 1
    new_page = _titrate(
        gen_titrate_q, x, distribution, q, step, var, data_rows, tol, 
        n_pages_left, **kwargs
    )
    # flag this as a titration page
    new_page.g = {'Titration': True}
    page.branch.pages.insert(page.index+1, new_page)


"""Data store"""

from ...app import db, socketio
from .data_frame import DataFrameType

import json
import pandas as pd
from sqlalchemy_mutable import MutableDictJSONType

from datetime import datetime

STATUS = ['Completed', 'InProgress', 'TimedOut']
DEFAULT_STATUS = {s: 0 for s in STATUS}


class DataStore(db.Model):
    """
    Each app has a unique data store.

    It contains two dataframes:

    1. `data`. The primary dataframe
    2. `meta`. Participant metadata, used to detect screenouts and duplicates.
    """
    id = db.Column(db.Integer, primary_key=True)
    _current_status = db.Column(MutableDictJSONType, default=DEFAULT_STATUS)
    data = db.Column(DataFrameType, default={})
    meta = db.Column(DataFrameType, default={})
    
    @property
    def current_status(self):
        """Add total Participants and time stamp to _current_status"""
        current_status = self._current_status.copy()
        current_status['Total'] = sum([current_status[s] for s in STATUS])
        return current_status
    
    def __init__(self):
        self.data, self.meta = {}, {}
        self.data.filename = 'data.csv'
        self.meta.filename = 'metadata.csv'
    
    def update_status(self, part):
        """Update current status
        
        Store Participant data on completion and time out.
        """
        if part._previous_status is not None:
            count = self._current_status[part._previous_status]
            self._current_status[part._previous_status] = max(count-1, 0)
        self._current_status[part.status] += 1
        current_status = json.dumps(self.current_status)
        socketio.emit('json', current_status, namespace='/participants-nsp')
    
    def store_participant(self, part):
        """Store data for given Participant"""
        self.remove_participant(part)
        self.data.append(
            part.cached_data if part.status in ('Completed', 'TimedOut')
            else part.get_data()
        )
        part.updated = False
        
    def remove_participant(self, part):
        """Remove data for given Participant"""
        id_var = self.data.get('ID') or []
        if not id_var or part.id not in id_var:
            return
        end = start = id_var.index(part.id)
        while end < len(id_var) and id_var[end] == part.id:
            end += 1
        self.data.remove(start, end)
        
    def to_pandas(self, dataframe='data'):
        """
        Parameters
        ----------
        dataframe : str
            Name of the dataframe to convert to pandas.

        Returns
        -------
        df : pandas.DataFrame
            Dataframe representation of the participant data.
        """
        return pd.DataFrame(getattr(self, dataframe) or {})


"""# Input"""

from .bases import InputBase
from ..app import db, settings
from ..functions.debug import send_datetime, send_keys
from ..models import Question

from selenium_tools import get_datetime

html_datetime_types = (
    'date',
    'datetime-local',
    'month',
    'time',
    'week',
)

def random_input(driver, question):
    """
    Default debug function for input questions. This function sends a random 
    string or number if the input takes text, or a random `datetime.datetime` 
    object if the input takes dates or times.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    question : hemlock.Input
    """
    if question.type in html_datetime_types:
        send_datetime(driver, question)
    else:
        send_keys(driver, question)

settings['Input'] = {
    'class': ['form-control'], 'type': 'text', 'debug': random_input
}


class Input(InputBase, Question):
    """
    Inputs take text input by default, or other types of html inputs.

    Inherits from [hemlock.models.InputBase`](bases.md) and 
    [`hemlock.Question`](../models/question.md).

    Parameters
    ----------
    label : str or None, default=None
        Input label.

    template : str, default='hemlock/input.html'
        Template for the input body.
    
    Examples
    --------
    ```python
    from hemlock import Input, Page, push_app_context

    app = push_app_context()

    Page(Input('<p>Input text here.</p>')).preview()
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'input'}

    def __init__(self, label=None, template='hemlock/input.html', **kwargs):
        super().__init__(label=label, template=template, **kwargs)

    def _record_data(self):
        if self.type in html_datetime_types:
            self.data = get_datetime(self.type, self.response) or None
        elif self.type == 'number' and self.response: 
            self.data = float(self.response)
        else:
            super()._record_data()
        return self


"""## Compile functions"""

from ..models import Compile, Page, ChoiceQuestion

from random import shuffle as shuffle_

@Compile.register
def call_method(obj, method_name, *args, **kwargs):
    """
    Calls one of the object's methods.

    Parameters
    ----------
    obj :
        Object whose methods will be called.

    method_name : str
        Names of the method to call.

    \*args, \*\*kwargs :
        Arguments and keyword arguments to pass to the method.

    Examples
    --------
    ```python
    from hemlock import Compile as C, Page, push_app_context

    app = push_app_context()

    p = Page(error='Error message', compile=C.call_method('clear_error'))
    p.preview()._compile().preview()
    ```
    """
    getattr(obj, method_name)(*args, **kwargs)

@Compile.register
def clear_error(obj):
    """
    Calls the object's `clear_error` method.

    Parameters
    ----------
    obj :
        Object whose `clear_error` method will be called.

    Examples
    --------
    ```python
    from hemlock import Compile as C, Page, push_app_context

    app = push_app_context()

    p = Page(error='Error message', compile=C.clear_error())
    p.preview()._compile().preview()
    ```
    """
    obj.clear_error()

@Compile.register
def clear_response(obj):
    """
    Calls the object's `clear_response` method.

    Parameters
    ----------
    obj :
        Object whose `clear_response` method will be called.

    Examples
    --------
    ```python
    from hemlock import Compile as C, Input, Page, push_app_context

    app = push_app_context()

    p = Page(Input(response='Hello World'), compile=C.clear_response())
    p.preview()._compile().preview()
    ```
    """
    obj.clear_response()

@Compile.register
def shuffle(obj, *attrs):
    """
    Shuffle an object's attributes.

    Parameters
    ----------
    obj :
        Objects whose attributes should be shuffled.

    \*attrs : str
        Names of attributes to shuffle.

    Notes
    -----
    If the object is a `hemlock.Page`, the default shuffled attribute is its 
    `questions`.

    If the object is a `hemlock.ChoiceQuestion`, the default shuffled
    attribute is its `choices`.

    Examples
    --------
    ```python
    from hemlock import Compile as C, Label, Page, push_app_context

    app = push_app_context()

    p = Page(
    \    *(Label('Label {}'.format(i)) for i in range(4)),
    \    compile=C.shuffle()
    )
    p.preview()._compile().preview()
    ```
    """
    if not attrs:
        if isinstance(obj, Page):
            attrs = ['questions']
        elif isinstance(obj, ChoiceQuestion):
            attrs = ['choices']
    [shuffle_(getattr(obj, attr)) for attr in attrs]


"""# Textarea"""

from ..app import db, settings
from ..functions.debug import send_keys
from ..models import Question
from .bases import InputBase

from flask import render_template

settings['Textarea'] = {'debug': send_keys, 'class': ['form-control'], 'rows': 3}


class Textarea(InputBase, Question):
    """
    Textareas provide large text boxes for free responses.

    Inherits from [`hemlock.qpolymorphs.InputBase`](bases.md) and 
    [`hemlock.Question`](../models/question.md).

    Parameters
    ----------
    label : str, default=None
        Textarea label.

    template : str, default='hemlock/textarea.html'
        Template for the textarea body.

    Notes
    -----
    Textareas have a default javascript which displays the character and word 
    count to participants. This will be appended to any `js` and `extra_js`
    arguments passed to the constructor.

    Examples
    --------
    ```python
    from hemlock import Page, Textarea, push_app_context

    app = push_app_context()

    Page(Textarea('<p>This is a textarea.</p>')).preview()
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'textarea'}

    _input_attr_names = [
        'class',
        'autofocus',
        'cols',
        'disabled',
        'maxlength',
        'placeholder',
        'readonly',
        'required',
        'rows',
        'wrap'
    ]

    def __init__(
            self, label=None, template='hemlock/textarea.html', **kwargs
        ):
        super().__init__(label=label, template=template, **kwargs)
        self.js.append(render_template('hemlock/textarea.js', q=self))


"""# Validation functions

These are built-in functions to validate a participant's response to a 
question. They return `None` if the response is valid, and an error message if 
the repsonse is invalid.
"""

from ..models import Validate
from ..tools.lang import join, plural
from .utils import (
    convert, correct_choices as correct_choices_, match as match_
)

import re
from operator import __ge__, __le__

# Require and type validation

RESP_TYPE_MSG = 'Please enter {}'

@Validate.register
def response_type(question, resp_type):
    """
    Validate that the response can be converted to a given type.
    
    Parameters
    ----------
    question : hemlock.Question

    resp_type : class
        The required type of response.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.response_type(float))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a number.
    ```
    """
    try:
        resp_type(question.response)
        return
    except:
        if resp_type == int:
            return RESP_TYPE_MSG.format('an integer')
        if resp_type == float:
            return RESP_TYPE_MSG.format('a number')
        return RESP_TYPE_MSG.format('the correct type of response')

REQUIRE_MSG = 'Please respond to this question'

@Validate.register
def require(question):
    """
    Require a response to this question.

    Parameters
    ----------
    question : hemlock.Question

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response=None, validate=V.require())
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please respond to this question.
    ```
    """
    return REQUIRE_MSG if question.response in (None, '') else None

# Set validation

IS_IN_MSG = 'Please enter {}'

@Validate.register
def is_in(question, valid_set, resp_type=None):
    """
    Validate that the question response is in a set of valid responses.

    Parameters
    ----------
    question : hemlock.Question
    
    valid_set : iterable
        Set of valid responses.

    resp_type : class or None, default=None
        Type of response expected; should match the type of elements in 
        `valid_set`.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='earth', validate=V.is_in(('wind', 'fire')))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter wind or fire.
    ```
    """
    resp, _ = convert(question.response, resp_type)
    if resp not in valid_set:
        return IS_IN_MSG.format(join('or', *valid_set))

NOT_IN_MSG = 'Please do not enter {}'

@Validate.register
def is_not_in(question, invalid_set, resp_type=None):
    """
    Validate that the question response is *not* in a set of invalid 
    responses.

    Parameters
    ----------
    question : hemlock.Question

    invalid_set : iterable
        Set of invalid responses.

    resp_type : class or None, default=None
        Type of response expected; should match the type of elements in
        `invalid_set`.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(
    \    response='earth', 
    \    validate=V.is_not_in(('earth', 'wind', 'fire'))
    )
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please do not enter earth, wind, or fire.
    ```
    """
    resp, _ = convert(question.response, resp_type)
    if resp in invalid_set:
        return NOT_IN_MSG.format(join('or', *invalid_set))

# Range validation

MAX_MSG = 'Please enter a response less than {}'

@Validate.register
def max_val(question, max_, resp_type=None):
    """
    Validate that the response does not exceed a maximum value.

    Parameters
    ----------
    question : hemlock.Question

    max_ : 
        Maximum value.

    resp_type : class or None, default=None
        Expected type of response. If `None`, the type of `max` will be used.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='101', validate=V.max_val(100))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a response less than 100.
    ```
    """
    error_msg = MAX_MSG.format(max_)
    return _compare_resp(question, max_, __le__, error_msg, resp_type)

MIN_MSG = 'Please enter a response greater than {}'

@Validate.register
def min_val(question, min_, resp_type=None):
    """
    Validate that the response does not deceed a minumum value.

    Parameters
    ----------
    question : hemlock.Question

    min_ : 
        Minimum value.

    resp_type : class or None, default=None
        Expected type of response. If `None`, the type of `min` will be used.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='-1', validate=V.min_val(0))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a response greater than 0.
    ```
    """
    error_msg = MIN_MSG.format(min_)
    return _compare_resp(question, min_, __ge__, error_msg, resp_type)

def _compare_resp(question, val, comparator, error_msg, resp_type=None):
    """
    Compare question response.

    Parameters
    ----------
    question : hemlock.Question

    val :
        Comparison value.

    comparator : callable
        Takes `question.resposne` and `val` and returns a bool indicating that 
        the comparison was 'successful'.

    error_msg : str
        Error message to be displayed if comparison is unsuccessful.

    resp_type : class or None, default=None
        Expected type of response. If `None`, the type of `val` will be used.
    """
    resp_type = resp_type or type(val)
    type_error_msg = response_type(question, resp_type)
    if type_error_msg is not None:
        return type_error_msg
    if not comparator(resp_type(question.response), val):
        return error_msg

RANGE_MSG = 'Please enter a response between {0} and {1}'

@Validate.register
def range_val(question, min_, max_, resp_type=None):
    """
    Validate that the response is in a given range.

    Parameters
    ----------
    question : hemlock.Question
    
    min_ : 
        Minimum value for the question response.

    max_ :
        Maximum value for the question response.

    resp_type : class or None, default=None
        Expected type of response. If `None`, the expected response type is 
        the type of `min` and `max`, which must be of the same type.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='101', validate=V.range_val(0, 100))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a response between 0 and 100.
    ```
    """
    if resp_type is None:
        assert type(min_) == type(max_)
        resp_type = type(min_)
    type_error_msg = response_type(question, resp_type)
    if type_error_msg is not None:
        return type_error_msg
    if not (min_ <= resp_type(question.response) <= max_):
        return RANGE_MSG.format(min_, max_)

# Length validation

EXACT_CHOICES_MSG = 'Please select exactly {0} {choice}'
EXACT_LEN_MSG = 'Please enter exactly {0} {character}'

@Validate.register
def exact_len(question, len_):
    """
    Validates the exact length of the repsonse. For a string response, this is 
    the length of the string. For a choices response, this is the number of 
    choices selected.

    Parameters
    ----------
    question : hemlock.Question

    len_ : int
        Required length of the response.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.exact_len(5))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter exactly 5 characters.
    ```
    """
    msg = require(question)
    if len_ and msg is not None:
        return msg
    if len(question.response) == len_:
        return
    if isinstance(question.response, list):
        return EXACT_CHOICES_MSG.format(len_, choice=plural(len_, 'choice'))
    return EXACT_LEN_MSG.format(len_, character=plural(len_, 'character'))

MAX_CHOICES_MSG = 'Please select at most {0} {choice}'
MAX_LEN_MSG = 'Please enter at most {0} {character}'

@Validate.register
def max_len(question, max_):
    """
    Validates the maximum length of the response. For a string response, this 
    is the length of the string. For a choices response, this is the number of 
    choices selected.

    Parameters
    ----------
    question : hemlock.Question

    max_ : int
        Maximum length of the response.

    Notes
    -----
    A response of `None` is assumed to satisfy the max length validation. Use 
    `Validate.require` to require a response that is not `None`.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.max_len(5))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter at most 5 characters.
    ```
    """
    if not question.response:
        return
    if len(question.response) <= max_:
        return
    if isinstance(question.response, list):
        return MAX_CHOICES_MSG.format(max_, choice=plural(max_, 'choice'))
    return MAX_LEN_MSG.format(max_, character=plural(max_, 'character'))

MIN_CHOICES_MSG = 'Please select at least {0} {choice}'
MIN_LEN_MSG = 'Please enter at least {0} {character}'

@Validate.register
def min_len(question, min_):
    """
    Valiadates the minimum length of the response. For a string response, this 
    is the length of the string. For a choices response, this is the number of 
    choices selected.

    Parameters
    ----------
    question : hemlock.Question
    
    min_ : int
        Minimum length of the response.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.min_len(15))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter at least 15 characters.
    ```
    """
    if min_ <= 0:
        return
    msg = require(question)
    if msg is not None:
        return msg
    if min_ <= len(question.response):
        return
    if isinstance(question.response, list):
        return MIN_CHOICES_MSG.format(min_, choice=plural(min_, 'choice'))
    return MIN_LEN_MSG.format(min_, character=plural(min_, 'character'))

RANGE_CHOICES_MSG = 'Please select between {0} and {1} choices'
RANGE_LEN_MSG = 'Please enter between {0} and {1} characters'

@Validate.register
def range_len(question, min_, max_):
    """
    Validates the range of the response length. For a string response, this is 
    the length of the string. For a choices response, this is the number of 
    choices selected.

    Parameters
    ----------
    question : hemlock.Question

    min_ : int
        Minimum response length.

    max_ : int
        Maximum response length.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.range_len(5, 10))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter 5 to 10 characters.
    ```
    """
    if min_ <= 0:
        return
    msg = require(question)
    if msg is not None:
        return msg
    if min_ <= len(question.response) <= max_:
        return
    if isinstance(question.response, list):
        return RANGE_CHOICES_MSG.format(min_, max_)
    return RANGE_LEN_MSG.format(min_, max_)

# Words validation

EXACT_WORDS_MSG = 'Please enter exactly {0} {word}'

@Validate.register
def exact_words(question, nwords):
    """
    Validate the exact number of words in the response.

    Parameters
    ----------
    question : hemlock.Question
    
    nwords : int
        Required number of words.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.exact_words(1))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter exactly 1 word.
    ```
    """
    msg = require(question)
    if nwords and msg is not None:
        # a response of `None` can satisfy `exact_words` if `value==0`
        return msg
    assert isinstance(question.response, str)
    if _num_words(question.response) != nwords:
        return EXACT_WORDS_MSG.format(nwords, word=plural(nwords, 'word'))

MAX_WORDS_MSG = 'Please enter at most {0} {word}'

@Validate.register
def max_words(question, max_):
    """
    Validates the maximum number of words in the response.

    Parameters
    ----------
    question : hemlock.Question

    max_ : int
        Maximum number of words.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.max_words(1))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter at most 1 word.
    ```
    """
    if not question.response:
        return
    assert isinstance(question.response, str)
    if _num_words(question.response) > max_:
        return MAX_WORDS_MSG.format(max_, word=plural(max_,'word'))

MIN_WORDS_MSG = 'Please enter at least {0} {word}'

@Validate.register
def min_words(question, min_):
    """
    Validates the minimum number of words in the repsonse.

    Parameters
    ----------
    question : hemlock.Question

    min_ : int
        Minimum number of words.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.min_words(3))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter at least 3 words.
    ```
    """
    if min_ <= 0:
        return
    msg = require(question)
    if msg is not None:
        return msg
    assert isinstance(question.response, str)
    if _num_words(question.response) < min_:
        return MIN_WORDS_MSG.format(min_, word=plural(min_,'word'))

RANGE_WORDS_MSG = 'Please enter between {0} and {1} words'

@Validate.register
def range_words(question, min_, max_):
    """
    Validates the number of words falls in a given range.

    Parameters
    ----------
    question : hemlock.Question

    min_ : int
        Minumum number of words.

    max_ : int
        Maximum number of words.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.range_words(3, 5))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter between 3 and 5 words.
    ```
    """
    if min_ <= 0:
        return
    msg = require(question)
    if msg is not None:
        return msg 
    assert isinstance(question.response, str)
    if not (min_ <= _num_words(question.response) <= max_):
        return RANGE_WORDS_MSG.format(min_, max_)

def _num_words(string):
    """Count the number of words in the string"""
    return len(re.findall(r'\w+', string))

# Decimal validation

EXACT_DECIMALS = 'Please enter a number with exactly {0} {decimal}'

@Validate.register
def exact_decimals(question, ndec):
    """
    Validates the exact number of decimals.

    Parameters
    ----------
    question : hemlock.Question

    ndec : int
        Required number of decimals.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='1', validate=V.exact_decimals(2))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a number with exactly 2 decimals.
    ```
    """
    msg, decimals = _get_decimals(question)
    if msg:
        return msg
    if decimals != ndec:
        return EXACT_DECIMALS.format(ndec, decimal=plural(ndec, 'decimal'))

MAX_DECIMALS = 'Please enter a number with at most {0} {decimal}'

@Validate.register
def max_decimals(question, max_):
    """
    Validates the maximum number of decimals.

    Parameters
    ----------
    question : hemlock.Question

    max_ : int
        Maximum number of decimals.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='1.123', validate=V.max_decimals(2))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a number with at most 2 decimals.
    ```
    """
    msg, decimals = _get_decimals(question)
    if msg:
        return msg
    if decimals > max_:
        return MAX_DECIMALS.format(max_, decimal=plural(max_, 'decimal'))

MIN_DECIMALS = 'Please enter a number with at least {0} {decimal}'

@Validate.register
def min_decimals(question, min_):
    """
    Validates the minumum number of decimals.

    Parameters
    ----------
    question : hemlock.Question

    min_ : int
        Minumum number of decimals.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='1', validate=V.min_decimals(2))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a number with at least 2 decimals.
    ```
    """
    msg, decimals = _get_decimals(question)
    if msg:
        return msg
    if decimals < min_:
        return MIN_DECIMALS.format(min_, decimal=plural(min_, 'decimal'))

RANGE_DECIMALS = 'Please enter a number with {0} to {1} decimals'

@Validate.register
def range_decimals(question, min_, max_):
    """
    Validates the number of decimals are in a given range.

    Parameters
    ----------
    question : hemlock.Question
    
    min_ : int
        Minimum number of decimals.

    max_ : int
        Maximum number of decimals.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='1.123', validate=V.range_decimals(0, 2))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a number with 0 to 2 decimals.
    ```
    """
    msg, decimals = _get_decimals(question)
    if msg:
        return msg
    if not (min_ <= decimals <= max_):
        return RANGE_DECIMALS.format(min_, max_)

def _get_decimals(question):
    """Return (error message, number of decimals) tuple

    This validation will fail if the response cannot be converted to `float`.
    """
    msg = response_type(question, float)
    if msg:
        return msg, None
    split = str(question.response).split('.')
    # number of decimals is 0 when no decimal point is specified.
    decimals = 0 if len(split) == 1 else len(split[-1])
    return None, decimals

# Regex validation

REGEX_MSG = 'Please enter a response with the correct pattern'

@Validate.register
def match(question, pattern):
    """
    Validate that the response matches the regex pattern.
    
    Parameters
    ----------
    question : hemlock.Question

    pattern : str or hemlock.Question
        Regex pattern to match. If this is a `Question`, the pattern is the 
        question's `response`'.

    Examples
    --------
    ```python
    from hemlock import Input, Validate as V, push_app_context

    app = push_app_context()

    inpt = Input(response='hello world', validate=V.match('goodbye *'))
    inpt._validate()
    inpt.error
    ```

    Out:

    ```
    Please enter a response with the correct pattern.
    ```
    """
    return REGEX_MSG if not match_(question, pattern) else None

# Choice validation

@Validate.register
def correct_choices(question, *values):
    """
    Validate that selected choice(s) is correct.
    
    Parameters
    ----------
    question : hemlock.Question

    \*values : 
        Values of the correct choices.

    Examples
    --------
    ```python
    from hemlock import Check, Validate as V, push_app_context

    app = push_app_context()

    check = Check(
    \    '<p>Select the correct choice.</p>',
    \    ['correct', 'incorrect', 'also incorrect'],
    \    validate=V.correct_choices('correct'),
    )
    check.response = check.choices[1]
    check._validate()
    check.error
    ```

    Out:

    ```
    Please select the correct choice.
    ```
    """
    if not correct_choices_(question, *values):
        if question.multiple:
            return '<p>Please select the correct choice(s).</p>'
        if len(values) == 1:
            return '<p>Please select the correct choice.</p>'
        return '<p>Please select one of the correct choices.</p>'


"""# Base for input questions"""

from ..app import db
from ..models import Question

from sqlalchemy_mutable import HTMLAttrsType


class InputBase():
    """
    This base class provides methods useful for debugging. Additionally, when 
    setting and getting attributes, it intercepts and automatically handles 
    input tag HTML attribues.
    """
    _input_attr_names = [
        'class',
        'type',
        'readonly',
        'disabled',
        'size',
        'maxlength',
        'max', 'min',
        'multiple',
        'pattern',
        'placeholder',
        'required',
        'step',
        'autofocus',
        'height', 'width',
        'list',
        'autocomplete',
    ]

    def __init__(self, *args, **kwargs):
        self.input_attrs = {}
        super().__init__(*args, **kwargs)

    def __getattribute__(self, key):
        if key == '_input_attr_names' or key not in self._input_attr_names:
            return super().__getattribute__(key)
        return self.input_attrs.get(key)

    def __setattr__(self, key, val):
        if key in self._input_attr_names:
            self.input_attrs[key] = val
        else:
            super().__setattr__(key, val)

    def input_from_driver(self, driver=None):
        """
        Parameters
        ----------
        driver : selenium.webdriver.chrome.webdriver.Webdriver
            Driver which will be used to select the input. Does not need to be Chrome.

        Returns
        -------
        input : selenium.webdriver.remote.webelement.WebElement
            Web element of the `<input>` tag associated with this model.
        """
        return driver.find_element_by_css_selector('#'+self.key)

    def label_from_driver(self, driver):
        """
        Parameters
        ----------
        driver : selenium.webdriver.chrome.webdriver.Webdriver
            Driver which will be used to select the label. Does not need to be Chrome.

        Returns
        -------
        label : selenium.webdriver.remote.webelement.WebElement
            Web element of the label tag associated with this model.
        """
        selector = 'label[for={}]'.format(self.key)
        return driver.find_element_by_css_selector(selector)


"""# Download button

See <https://dsbowen.github.io/flask-download-btn/> for more details.
"""

from ..app import db
from ..models import Question

from flask import render_template
from flask_download_btn import DownloadBtnManager, DownloadBtnMixin


@DownloadBtnManager.register
class Download(DownloadBtnMixin, Question):
    """
    Allows participants to download files.

    Inherits from 
    [`flask_download_btn.DownloadBtnMixin`](https://dsbowen.github.io/flask-download-btn/download_btn_mixin/) and 
    [`hemlock.Question`](../models/question.md).

    Parameters
    ----------
    label : str or None, default=None
        Download button label.

    template : str, default='hemlock/download.html'
        Download button body template.

    Examples
    --------
    ```python
    from hemlock import Download, Page, push_app_context

    app = push_app_context()

    Page(Download(
    \    '<p>Click here to download a file.</p>',
    \    downloads=('HELLO_WORLD_URL', 'hello_world.txt')
    )).preview()
    ```

    Replace `'HELLO_WORLD_URL'` with your file download URL. Note that the 
    download button will not download your file from a preview.
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'download'}

    def __init__(
            self, label=None, template='hemlock/download.html', **kwargs
        ):
        super().__init__(label=label, template=template, **kwargs)

    def _render_js(self):
        return self.js.render() + self.render_script()


"""# Debug functions

Debug functions tell the AI participant what to do during debugging. They 
generally take a selenium webdriver as their first argument and a page or 
question as their second argument.

Notes
-----
The following examples open a webdriver. After running the examples, close the
driver with `driver.close()`.

By default, the last debug function of a page navigates. To remove this, run
`page.debug.pop()`.
"""

from ..app import settings
from ..models import Debug
from .utils import random_datetime, random_num, random_str

from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from selenium_tools import (
    drag_range as drag_range_, send_datetime as send_datetime_
)

import math
from datetime import datetime, timedelta
from random import randint, random, randrange, shuffle

from time import sleep

def _is_displayed(element, max_wait=1, wait_interval=1):
    """
    Parameters
    ----------
    element : selenium.webdriver.remote.webelement.WebElement
        Checks if this element is displayed.

    max_wait : float, default=1
        Maximum number of seconds to wait for element to be displayed.

    wait_interval : float, default=1
        Interval in between checking to see if element is displayed.

    Returns
    -------
    is_displayed : bool
        Indicates that the web element is displayed.
    """
    for _ in range(0, max_wait, wait_interval):
        if element.is_displayed():
            return True
        sleep(wait_interval)
    return False

# Page debugging
@Debug.register
def forward(driver, page, max_wait=30, wait_interval=3):
    """
    Click the forward button.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    page : hemlock.Page

    Examples
    --------
    ```python
    from hemlock import Debug as D, Page, push_app_context
    from hemlock.tools import chromedriver

    app = push_app_context()

    driver = chromedriver()

    p = Page(debug=[D.debug_questions(), D.forward()])
    p.preview(driver)._debug(driver)
    ```
    """
    forward_btn = driver.find_element_by_id('forward-btn')
    if _is_displayed(forward_btn, max_wait, wait_interval):
        forward_btn.click()
    else:
        raise TimeoutError('Forward button not displayed')

@Debug.register
def back(driver, page, max_wait=30, wait_interval=1):
    """
    Click the back button.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    page : hemlock.Page

    Examples
    --------
    ```python
    from hemlock import Debug as D, Page, push_app_context
    from hemlock.tools import chromedriver

    app = push_app_context()

    driver = chromedriver()

    p = Page(debug=[D.debug_questions(), D.back()])
    p.preview(driver)._debug(driver)
    ```
    """
    back_btn = driver.find_element_by_id('back-btn')
    if _is_displayed(back_btn, max_wait, wait_interval):
        back_btn.click()
    else:
        raise TimeoutError('Back button not displayed')

# Textarea and input debugging

@Debug.register
def send_keys(driver, question, *keys, p_num=.5):
    """
    Send the specified keys to the `<textarea>` or `<input>`.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    question : hemlock.Question

    \*keys : 
        Keys to send to the textarea or input. If empty, keys are randomly 
        selected.

    p_num : float, default=.5
        Probability of sending a random number if keys are not specified (as 
        opposed to a random string).

    Notes
    -----
    This debug function is skipped if the question is not displayed.

    Examples
    --------
    ```python
    from hemlock import Debug as D, Input, Page, push_app_context
    from hemlock.tools import chromedriver

    app = push_app_context()

    driver = chromedriver()

    p = Page(Input(debug=D.send_keys('hello world')))
    p.preview(driver)._debug(driver)
    ```
    """
    def random_keys():
        html_datetime_types = (
            'date', 'datetime-local', 'month', 'time', 'week'
        )
        type_ = inpt.get_attribute('type')
        if inpt.tag_name == 'textarea' or not type_ or type_ == 'text':
            if random() < p_num:
                random_num(inpt)
            else:
                random_str(inpt)
        elif type_ == 'number':
            random_num(inpt)
        elif type_ in html_datetime_types:
            random_datetime(inpt)

    inpt = question.input_from_driver(driver)
    if _is_displayed(inpt):
        inpt.clear()
        if keys:
            [inpt.send_keys(key) for key in keys]
        else:
            random_keys()

# Date and time input
@Debug.register
def send_datetime(driver, question, datetime_=None):
    """
    Send a `datetime.datetime` object to an input. Inputs should be of type
    'date', 'datetime-local', 'month', 'time', or 'week',

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    question : hemlock.Question

    datetime_ : datetime.datetime or None, default=None
        The datetime object to send. If `None`, a date and time are chosen
        randomly.

    Notes
    -----
    This debug function is skipped if the question is not displayed.

    Examples
    --------
    ```python
    from hemlock import Debug as D, Input, Page, push_app_context
    from hemlock.tools import chromedriver

    from datetime import datetime

    app = push_app_context()

    driver = chromedriver()

    p = Page(Input(type='date', debug=D.send_datetime(datetime.utcnow())))
    p.preview(driver)._debug(driver)
    ```
    """
    inpt = question.input_from_driver(driver)
    if _is_displayed(inpt):
        inpt.clear()
        random_datetime(inpt)

# Range debugger

@Debug.register
def drag_range(driver, range_, target=None, tol=0, max_iter=10):
    """
    Drag a range slider to specified target value.
    
    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    range_ : hemlock.Range

    target : float or None, default=None
        Target value to which the slider should be dragged. If `None`, a
        random target value will be chosen.

    tol : float, default=0
        Tolerance for error if the slider cannot be dragged to the exact
        target.

    max_iter : int, default=10
        Maximum number of iterations for the slider to reach the target.

    Notes
    -----
    This debug function is skipped if the question is not displayed.

    Examples
    --------
    ```python
    from hemlock import Debug as D, Page, Range, push_app_context
    from hemlock.tools import chromedriver

    app = push_app_context()

    driver = chromedriver()

    p = Page(Range(debug=D.drag_range(80)))
    p.preview(driver)._debug(driver)
    ```
    """
    if target is None:
        target = randint(range_.min, range_.max)
        target = round(target / range_.step) * range_.step
    inpt = range_.input_from_driver(driver)
    if _is_displayed(inpt):
        drag_range_(
            driver, inpt, target,
            horizontal=True, # all sliders are horizontal for now
            tol=tol,
            max_iter=max_iter
        )

# Choice question debugger

@Debug.register
def click_choices(driver, question, *values, if_selected=None, max_clicks=5):
    """
    Click on choices or options.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    question : hemlock.ChoiceQuestion
    
    \*values : 
        Values of the choices on which to click. If no choices are specified, 
        the debugger will click on random choices.

    if_selected : bool or None, default=None
        Indicates that the choices will be clicked only if they are already 
        selected. If `False` the choices will be clicked only if they are not 
        already selected. If `None` the choices will be clicked whether or 
        not they are selected.

    Notes
    -----
    Will not attempt to click choices or options which are not displayed.

    Examples
    --------
    ```python
    from hemlock import Binary, Debug as D, Page, push_app_context
    from hemlock.tools import chromedriver

    app = push_app_context()

    driver = chromedriver()

    p = Page(
    \    Binary(
    \        '<p>Click "Yes".</p>', 
    \        debug=D.click_choices('Yes')
    \    )
    )
    p.preview(driver)._debug(driver)
    ```
    """
    def choose_values():
        order = list(range(len(question.choices)))
        shuffle(order)
        n_values = min(randint(0, len(question.choices)), max_clicks)
        return [question.choices[i].value for i in order[:n_values]]

    values = values if values else choose_values()
    for c in question.choices:
        if c.value in values and c.is_displayed(driver):
            c.click(driver, if_selected=if_selected)

@Debug.register
def clear_choices(driver, question):
    """
    Clear selected choices.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver

    question : hemlock.ChoiceQuestion

    Notes
    -----
    Intended only for questions in which multiple choices may be selected.

    Examples
    --------
    ```python
    from hemlock import Check, Debug as D, Page, push_app_context
    from hemlock.tools import chromedriver

    app = push_app_context()

    driver = chromedriver()

    p = Page(
    \    Check(
    \        "<p>Which ice cream flavors do you like?</p>",
    \        ['Chocolate', 'Vanilla', 'Strawberry'],
    \        default='Chocolate', 
    \        multiple=True,
    \        debug=D.clear_choices()
    \    )
    )
    p.preview(driver)._debug(driver)
    ```
    """
    if not question.choices:
        return
    if not question.multiple:
        print("Warning: Only multiple choice questions cannot be cleared")
        return
    for c in question.choices:
        if c.is_displayed(driver):
            c.click(driver, if_selected=True)


"""# Check"""

from ..app import db, settings
from ..functions.debug import click_choices
from ..models import ChoiceQuestion
from .bases import InputBase
from .choice import ChoiceListType


class CheckBase(InputBase, ChoiceQuestion):
    """
    Base for `hemlock.Binary` and `hemlock.Check` question types. Inherits 
    from [`hemlock.ChoiceQuestion`](../models/question.md).

    Attributes
    ----------
    align : str, default='left'
        Alignment of the choice text. Value can be `'left'`, `'center'`, 
        or `'right'`.

    inline : bool, default=False
        Indicates that choices should be 
        [inline](https://getbootstrap.com/docs/4.0/components/forms/#inline), 
        as opposed to vertical.

    Notes
    -----
    A choice can be input as any of the following:

    1. `Choice` object.
    2. `str`, treated as the choice label.
    3. `(choice label, value)` tuple.
    4. `(choice label, value, name)` tuple.
    5. Dictionary with choice keyword arguments.
    """
    choices = db.Column(ChoiceListType)
    multiple = db.Column(db.Boolean, default=False)
    inline = db.Column(db.Boolean, default=False)


settings['Check'] = {
    'inline': False,
    'debug': click_choices,
    'multiple': False,
}


class Check(CheckBase):
    """
    Check questions use radio inputs if only one choice can be selected, or 
    checkbox inputs if multiple choices can be selected.

    Inherits from [`hemlock.qpolymorphs.check.CheckBase`](check.md).

    Its default debug function is 
    [`click_choices`](../functions/debug.md#hemlockfunctionsdebugclick_choices).

    Parameters
    ----------
    label : str or None, default=None
        Check question label.

    choices : list of hemlock.Choice, str, tuple, or dict, default=[]
        Choices which participants can check.

    template : str, default='hemlock/check.html'
        File name of the Jinja template.

    Attributes
    ----------
    choices : list of `hemlock.Choice`
        Set from the `choices` parameter.

    inline : default=False

    multiple : bool, default=False
        Indicates that the participant may select multiple choices.

    Examples
    --------
    ```python
    from hemlock import Check, Page, push_app_context

    app = push_app_context()

    Page(Check('<p>Check one.</p>', ['Yes','No','Maybe'])).preview()
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'check'}

    def __init__(
            self, label=None, choices=[], template='hemlock/check.html', 
            **kwargs
        ):
        super().__init__(label, choices, template, **kwargs)


settings['Binary'] = {
    'inline': True,
    'debug': click_choices,
    'multiple': False,
    'record_choice_index': False,
}


class Binary(CheckBase):
    """
    Binary question use radio inputs and have two options, coded as 0 and 1.

    Inherits from [`hemlock.qpolymorphs.check.CheckBase`](check.md).

    Its default debug function is 
    [`click_choices`](../functions/debug.md#hemlockfunctionsdebugclick_choices).
    
    Parameters
    ----------
    label : str or None, default=None
        Check question label.

    choices : list of [str, str], default=['Yes', 'No']
        Choices which participants can check. The first choice is coded as 1,
        the second as 0.

    template : str, default='hemlock/check.html'
        File name of the Jinja template.

    Attributes
    ----------
    choices : list of `hemlock.Choice`
        Set from the `choices` parameter.
        
    inline : bool, default=True

    multiple : bool, default=False
        Indicates that the participant may select multiple choices.

    Examples
    --------
    ```python
    from hemlock import Binary, Page, push_app_context

    app = push_app_context()

    Page(Binary('<p>Yes or no?</p>')).preview()
    ```
    """
    id = db.Column(db.Integer, db.ForeignKey('question.id'), primary_key=True)
    __mapper_args__ = {'polymorphic_identity': 'binary'}

    def __init__(
            self, label=None, choices=['Yes', 'No'], 
            template='hemlock/check.html', **kwargs
        ):
        assert len(choices) == 2, 'Binary question require 2 choices'
        choices = [(choices[0], 1), (choices[1], 0)]
        super().__init__(label, choices, template, **kwargs)